# -*- coding: utf-8; -*-
"""
Copyright (c) 2019 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

import gc
import platform
import sys
from os import listdir, rename, remove
from os.path import splitext, join, dirname

import psutil
from PyQt5 import QtCore
from numpy import uint16, uint8

from align_frames import AlignFrames
from alignment_points import AlignmentPoints
from configuration import PostprocDataObject
from exceptions import NotSupportedError, InternalError, ArgumentError, Error
from frames import Frames, Calibration
from miscellaneous import Miscellaneous
from rank_frames import RankFrames
from stack_frames import StackFrames
from timer import timer

# The following lists define the allowed file extensions for still images.
image_extensions = ['.tif', '.tiff', '.fit', '.fits', '.jpg', '.png']

class Workflow(QtCore.QObject):
    master_dark_created_signal = QtCore.pyqtSignal(bool)
    master_flat_created_signal = QtCore.pyqtSignal(bool)
    work_next_task_signal = QtCore.pyqtSignal(str)
    report_error_signal = QtCore.pyqtSignal(str)
    abort_job_signal = QtCore.pyqtSignal(str)
    work_current_progress_signal = QtCore.pyqtSignal(str, int)
    set_main_gui_busy_signal = QtCore.pyqtSignal(bool)
    set_status_bar_signal = QtCore.pyqtSignal(str, str)
    create_image_window_signal = QtCore.pyqtSignal()
    update_image_window_signal = QtCore.pyqtSignal(object)
    terminate_image_window_signal = QtCore.pyqtSignal()

    def __init__(self, main_gui):
        super(Workflow, self).__init__()
        self.main_gui = main_gui
        self.configuration = main_gui.configuration

        self.my_timer = None

        self.frames = None
        self.rank_frames = None
        self.align_frames = None
        self.alignment_points = None
        self.stack_frames = None
        self.stacked_image_name = None
        self.postprocessed_image_name = None
        self.postprocessed_image = None
        self.postproc_input_image = None
        self.postproc_input_name = None
        self.activity = None
        self.attached_log_name = None
        self.attached_log_name_new = None
        self.attached_log_file = None
        self.stdout_saved = None
        self.output_redirected = False
        self.protocol_file = None

        # Switch alignment point debugging on / off.
        self.debug_AP = False

        # Print info on the platform used.
        platform_name = platform.system()
        processor_name = platform.processor()

        python_dir = dirname(sys.executable)
        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol(
                "Operating system used: " + platform_name + ", Processor used: " + processor_name +
                "\n           Python interpreter location: " +
                python_dir, self.attached_log_file, precede_with_timestamp=True)

        # Check if the configuration was imported from an older version. If so, print a message.
        # Please note that the "version imported from" parameter is set only if a configuration
        # has been read.
        if self.configuration.global_parameters_protocol_level > 1 and \
                self.configuration.configuration_read and \
                self.configuration.global_parameters_version_imported_from != \
                self.configuration.global_parameters_version:
            Miscellaneous.protocol("           Configuration imported from older version: " +
                                   self.configuration.global_parameters_version_imported_from,
                                   self.attached_log_file, precede_with_timestamp=False)

        # Create the calibration object, used for potential flat / dark corrections.
        self.calibration = Calibration(self.configuration)

    @QtCore.pyqtSlot(list)
    def execute_create_master_dark(self, dark_names):
        # Create a new master dark.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Creating a new master dark frame +++",
                                   self.attached_log_file, precede_with_timestamp=True)
        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol("           Input frames: " + dark_names[0],
                                   self.attached_log_file, precede_with_timestamp=False)

        try:
            self.set_main_gui_busy_signal.emit(True)
            self.calibration.create_master_dark(dark_names[0])
            if self.calibration.warn_message is not None and \
                    self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("Warning in creating a new master dark frame: " +
                                       self.calibration.warn_message,
                                       self.attached_log_file, precede_with_timestamp=True)
            self.master_dark_created_signal.emit(True)
        except Exception as e:
            if self.configuration.global_parameters_protocol_level > 0:
                self.report_error_signal.emit("Error in creating master dark frame: " + str(e))
            self.master_dark_created_signal.emit(False)

    @QtCore.pyqtSlot(list)
    def execute_create_master_flat(self, flat_names):

        # Create a new master flat.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Creating a new master flat frame +++",
                                   self.attached_log_file, precede_with_timestamp=True)
        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol("           Input frames: " + flat_names[0],
                                   self.attached_log_file, precede_with_timestamp=False)

        try:
            self.set_main_gui_busy_signal.emit(True)
            self.calibration.create_master_flat(flat_names[0])
            if self.calibration.warn_message is not None and \
                    self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("Warning in creating a new master flat frame: " +
                                       self.calibration.warn_message,
                                       self.attached_log_file, precede_with_timestamp=True)
            self.master_flat_created_signal.emit(True)
        except Error as e:
            if self.configuration.global_parameters_protocol_level > 0:
                self.report_error_signal.emit("Error in creating master flat frame: " +
                                              str(e) + ", flat frame calibration de-activated")
            self.master_flat_created_signal.emit(False)

    @QtCore.pyqtSlot()
    def execute_reset_masters(self):

        # De-activate master frames.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ De-activating master frames +++", self.attached_log_file,
                                   precede_with_timestamp=True)
        self.calibration.reset_masters()

    @QtCore.pyqtSlot(object)
    def execute_frames(self, job):
        self.job = job

        # Reset the potential new log file name (required if parameters are to be encoded in file
        # name).
        self.attached_log_name_new = None

        # Remove objects from previous job to clean up RAM.
        for obj in [self.frames, self.rank_frames, self.align_frames, self.alignment_points,
                    self.stack_frames]:
            if obj is not None:
                del obj
        self.frames = None
        self.rank_frames = None
        self.align_frames = None
        self.alignment_points = None
        self.stack_frames = None

        # Force the garbage collector to release unreferenced objects.
        gc.collect()

        # Update the status bar in the main GUI.
        self.input_name = self.job.file_name
        self.set_status_bar_processing_phase("reading frames")

        # A job can either "stack" images or "postprocess" a single image. In the latter case,
        # input is a single image file.
        #
        # Images for stacking can either be extracted from a video file or a batch of single
        # photographs. In the first case, input_type is set to 'video', in the second case to
        # 'image'.

        if self.job.type == 'postproc':
            self.activity = 'postproc'
            self.postproc_input_name = self.job.name

            # Reset the postprocessed image to None. This way, in saving the postprocessing result,
            # it can be checked if an image was computed in the workflow thread.
            self.postprocessed_image = None
            self.postprocessed_image_name = PostprocDataObject.set_file_name_processed(
                self.job.name, self.configuration.postproc_suffix,
                self.configuration.global_parameters_image_format)
            self.attached_log_name = splitext(self.job.name)[0] + '_postproc-log.txt'

        # For video file input, the Frames constructor expects the video file name for "names".
        elif self.job.type == 'video':
            self.activity = 'stacking'
            names = self.job.name
            self.attached_log_name = splitext(self.job.name)[0] + '_stacking-log.txt'

        # For single image input, the Frames constructor expects a list of image file names for
        # "names".
        else:  # input_type = 'image'
            self.activity = 'stacking'
            # Include only names with image extensions.
            names = [join(self.job.name, name) for name in listdir(self.job.name) if
                     splitext(name)[-1].lower() in image_extensions]
            self.attached_log_name = self.job.name + '_stacking-log.txt'

        # Redirect stdout to a file if requested.
        if self.configuration.global_parameters_write_protocol_to_file != self.output_redirected:
            # Output currently redirected. Reset to stdout.
            if self.output_redirected:
                sys.stdout = self.stdout_saved
                self.output_redirected = False
            # Currently set to stdout, redirect to file now.
            else:
                try:
                    self.stdout_saved = sys.stdout
                    sys.stdout = open(self.configuration.protocol_filename, 'a+')
                    self.output_redirected = True
                except IOError:
                    pass

        # Create logfile if requested to store the log with the stacked file.
        if self.attached_log_file:
            self.attached_log_file.close()
        if self.configuration.global_parameters_store_protocol_with_result:
            self.attached_log_file = open(self.attached_log_name, "w+")
        else:
            self.attached_log_file = None

        # Write a header to stdout and optionally to the logfile.
        if self.configuration.global_parameters_protocol_level > 0:
            decorator_line = (len(self.job.name) + 28) * "*"
            Miscellaneous.protocol(decorator_line, self.attached_log_file,
                                   precede_with_timestamp=False)
            Miscellaneous.protocol("Start processing " + self.job.name, self.attached_log_file)
            Miscellaneous.protocol("           Software version used: " +
                                   self.configuration.global_parameters_version,
                                   self.attached_log_file, precede_with_timestamp=False)
            Miscellaneous.protocol(decorator_line, self.attached_log_file,
                                   precede_with_timestamp=False)

        # Initalize the timer object used to measure execution times of program sections.
        self.my_timer = timer()
        self.my_timer.create('Execution over all')

        if self.activity == 'stacking':

            # Write parameters to the protocol.
            if self.configuration.global_parameters_protocol_level > 1:
                Miscellaneous.print_stacking_parameters(self.configuration, self.attached_log_file)

            try:
                if self.configuration.global_parameters_maximum_memory_active:
                    available_ram = float(self.configuration.global_parameters_maximum_memory_amount)
                else:
                    # Look up the available RAM (without paging)
                    virtual_memory = dict(psutil.virtual_memory()._asdict())
                    available_ram = virtual_memory['available'] / 1e9

                self.frames = Frames(self.configuration, names, type=self.job.type,
                                     bayer_option_selected=self.job.bayer_option_selected,
                                     calibration=self.calibration,
                                     progress_signal=self.work_current_progress_signal)

                # If buffering is not automatic, set the buffering_level as requested by the user.
                if self.configuration.global_parameters_buffering_level != -1:
                    # Decide on the objects to be buffered, depending on configuration parameter.
                    buffering_level_set = self.configuration.global_parameters_buffering_level
                    # Compute the approximate RAM usage of this job at the selected buffering level.
                    needed_ram = self.frames.compute_required_buffer_size(buffering_level_set)
                else:
                    needed_ram = None

                # If buffering was set to "auto", compute the highest possible value. If the
                # buffering level requested explicitly was too high, test if lowering the
                # buffering level would help.
                if self.configuration.global_parameters_buffering_level == -1 or needed_ram > available_ram:
                    buffering_level_set = None
                    for level in range(4, -1, -1):
                        alternative_ram = self.frames.compute_required_buffer_size(level)
                        if alternative_ram < available_ram:
                            buffering_level_set = level
                            needed_ram = alternative_ram
                            break

                    # Check if the job can be processed with the requested buffering level.
                    message = None
                    if buffering_level_set is None:
                        message = "Error: Too little RAM for this job, continuing with the next one"

                    # If an appropriate level other then the one set by the user was found, write it
                    # as a recommendation to the protocol. The job is aborted anyway.
                    elif self.configuration.global_parameters_buffering_level != -1:
                        message = "Error: Too little RAM for chosen buffering level, recommended " \
                                  "level: " + str(
                            buffering_level_set) + ", continuing with next job"

                    # Buffering is set to "auto" and an appropriate buffering level was found. Set
                    # the buffering in the frames object.
                    else:
                        self.frames.set_buffering(buffering_level_set)

                    # The job does not fit in RAM, continue with the next job.
                    if message:
                        self.abort_job_signal.emit(message)
                        return

                # The user has set the buffering level explicitly, and the job fits in memory. Set
                # the buffering in the frames object.
                else:
                    self.frames.set_buffering(buffering_level_set)

                if self.configuration.global_parameters_protocol_level > 1:
                    Miscellaneous.protocol("+++ Buffering level is " +
                                           str(buffering_level_set) +
                                           " +++", self.attached_log_file)
                    Miscellaneous.protocol(
                        "           RAM required (Gbytes): " + str(round(needed_ram, 2)) +
                        ", available: " + str(round(available_ram, 2)), self.attached_log_file,
                        precede_with_timestamp=False)

                if self.configuration.global_parameters_protocol_level > 0:
                    Miscellaneous.protocol("+++ Start reading frames +++", self.attached_log_file)

                if self.frames.warn_message is not None and \
                    self.configuration.global_parameters_protocol_level > 0:
                    Miscellaneous.protocol(
                        "Warning in opening SER file: " + self.frames.warn_message,
                        self.attached_log_file, precede_with_timestamp=True)
                if self.configuration.global_parameters_protocol_level > 1:
                    Miscellaneous.protocol(
                        "           Number of frames: " + str(self.frames.number) +
                        ", image shape: " + str(self.frames.shape), self.attached_log_file,
                        precede_with_timestamp=False)
                    if self.job.bayer_option_selected == 'Auto detect color':
                        Miscellaneous.protocol(
                            "           Debayer pattern detected automatically: '" +
                            self.frames.bayer_pattern + "'",
                            self.attached_log_file, precede_with_timestamp=False)
                        self.job.bayer_pattern = self.frames.bayer_pattern
                    else:
                        Miscellaneous.protocol(
                            "           Debayer pattern selected manually: '" +
                            self.job.bayer_option_selected + "'",
                            self.attached_log_file, precede_with_timestamp=False)
                    if self.frames.dt0 == 'uint16':
                        dynamic_range = '16 bit'
                    elif self.frames.dt0 == 'uint8':
                        dynamic_range = '8 bit'
                    else:
                        dynamic_range = 'undefined'
                    Miscellaneous.protocol(
                        "           Dynamic range of input frames: " + dynamic_range + ".",
                        self.attached_log_file, precede_with_timestamp=False)
                    if self.frames.shift_pixels:
                        Miscellaneous.protocol(
                            "           Pixel values are multiplied by 2**" + str(
                                self.frames.shift_pixels) + " to use the full dynamic range of 16 "
                                                            "bits.",
                            self.attached_log_file, precede_with_timestamp=False)
                    if self.frames.calibration_matches:
                        if self.calibration.master_dark_frame_adapted is not None and \
                                self.calibration.inverse_master_flat_frame is not None:
                            Miscellaneous.protocol(
                                "           Dark / flat frame calibration is active",
                                self.attached_log_file, precede_with_timestamp=False)
                        elif self.calibration.master_dark_frame_adapted is not None:
                            Miscellaneous.protocol("           Dark frame calibration is active",
                                                   self.attached_log_file,
                                                   precede_with_timestamp=False)
                        elif self.calibration.inverse_master_flat_frame is not None:
                            Miscellaneous.protocol("           Flat frame calibration is active",
                                                   self.attached_log_file,
                                                   precede_with_timestamp=False)
                    else:
                        Miscellaneous.protocol(
                            "           No matching master dark / flat frames found, "
                            "calibration de-activated",
                            self.attached_log_file, precede_with_timestamp=False)
            except Error as e:
                self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
                return
            except Exception as e:
                self.abort_job_signal.emit(
                    "Error in opening/reading frames: " + str(e) + ", continuing with next job")
                return



            # The RAM seems to be sufficient, continue with ranking frames.
            self.work_next_task_signal.emit("Rank frames")

        # Job type is 'postproc'.
        else:
            try:
                self.postproc_input_image = Frames.read_image(self.postproc_input_name)
            except Error as e:
                self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
                return
            except Exception as e:
                self.abort_job_signal.emit(
                    "Error in reading image file: " + str(e) + ", continuing with next job")
                return

            # Convert 8 bit to 16 bit.
            if self.postproc_input_image.dtype == uint8:
                self.postproc_input_image = self.postproc_input_image.astype(uint16) * 256
            self.work_next_task_signal.emit("Postprocessing")

    @QtCore.pyqtSlot()
    def execute_rank_frames(self):

        # Reset the frame index translation.
        self.frames.reset_index_translation()

        self.set_status_bar_processing_phase("ranking frames")
        # Rank the frames by their overall local contrast.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start ranking frames +++", self.attached_log_file)
        self.my_timer.create_no_check('Ranking frames')

        try:
            self.rank_frames = RankFrames(self.frames, self.configuration,
                                          self.work_current_progress_signal)
            self.rank_frames.frame_score()
            self.my_timer.stop('Ranking frames')
        except Error as e:
            self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
            self.my_timer.stop('Ranking frames')
            return
        except Exception as e:
            self.abort_job_signal.emit(
                "Error: " + str(e) + ", continuing with next job")
            self.my_timer.stop('Ranking frames')
            return

        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol(
                "           Index of best frame: " + str(self.rank_frames.frame_ranks_max_index + 1),
                self.attached_log_file, precede_with_timestamp=False)

        self.work_next_task_signal.emit("Select frames")

    @QtCore.pyqtSlot()
    def execute_set_index_translation_table(self):
        # If in the frame selection dialog the status of at least one frame was changed, update
        # the index translation table.
        if not all(self.frames.index_included):
            self.frames.set_index_translation()
            self.rank_frames.set_index_translation(self.frames.index_translation)
        else:
            self.frames.reset_index_translation()
            self.rank_frames.reset_index_translation()

        self.work_next_task_signal.emit("Align frames")

    @QtCore.pyqtSlot(int, int, int, int)
    def execute_align_frames(self, y_low_opt, y_high_opt, x_low_opt, x_high_opt):

        self.set_status_bar_processing_phase("aligning frames")
        # Initialize the frame alignment object.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Initializing frame alignment +++", self.attached_log_file)
        self.align_frames = AlignFrames(self.frames, self.rank_frames, self.configuration,
                                        progress_signal=self.work_current_progress_signal)

        if self.configuration.align_frames_mode == "Surface":

            auto_execution = False
            if y_low_opt == 0 and y_high_opt == 0 and x_low_opt == 0 and x_high_opt == 0:
                auto_execution = True
            elif (y_high_opt - y_low_opt) / self.frames.shape[
                0] > self.configuration.align_frames_max_stabilization_patch_fraction or (
                    x_high_opt - x_low_opt) / self.frames.shape[
                1] > self.configuration.align_frames_max_stabilization_patch_fraction and \
                    self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("           Stabilization patch selected manually is "
                                       "too large, switch to automatic mode",
                                       self.attached_log_file, precede_with_timestamp=False)
                auto_execution = True
            elif (y_high_opt - y_low_opt) / self.frames.shape[
                0] < self.configuration.align_frames_min_stabilization_patch_fraction or (
                    x_high_opt - x_low_opt) / self.frames.shape[
                1] < self.configuration.align_frames_min_stabilization_patch_fraction and \
                    self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("           Stabilization patch selected manually is "
                                       "too small, switch to automatic mode",
                                       self.attached_log_file, precede_with_timestamp=False)
                auto_execution = True

            # Compute the local rectangular patch in the image where the L gradient is highest
            # in both x and y direction. The scale factor specifies how much smaller the patch
            # is compared to the whole image frame. In batch mode, variable "auto_execution" is
            # set to "True", and the automatic patch computation is the only option.
            if auto_execution or self.configuration.align_frames_automation:

                self.my_timer.create_no_check('Select optimal alignment patch')
                try:
                    (y_low_opt, y_high_opt, x_low_opt,
                     x_high_opt) = self.align_frames.compute_alignment_rect(
                        self.configuration.align_frames_rectangle_scale_factor)
                except (ArgumentError) as e:
                    self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
                    self.my_timer.stop('Select optimal alignment patch')
                    return
                self.my_timer.stop('Select optimal alignment patch')
                if self.configuration.global_parameters_protocol_level > 1:
                    Miscellaneous.protocol(
                        "           Alignment rectangle, computed automatically: " + str(
                            y_low_opt) + "<y<" + str(y_high_opt) + ", " + str(
                            x_low_opt) + "<x<" + str(x_high_opt), self.attached_log_file,
                        precede_with_timestamp=False)

            # As an alternative, set the coordinates of the rectangular patch explicitly.
            else:
                # The image displayed in the stabilization patch editor was shrunk on all four
                # sides by a number of pixels given by the alignment search width parameter.
                # Therefore, the resulting coordinates of the stabilization patch have to be
                # corrected by this offset now.
                y_low_opt += self.configuration.align_frames_search_width
                y_high_opt += self.configuration.align_frames_search_width
                x_low_opt += self.configuration.align_frames_search_width
                x_high_opt += self.configuration.align_frames_search_width

                self.align_frames.set_alignment_rect(y_low_opt, y_high_opt, x_low_opt, x_high_opt)
                if self.configuration.global_parameters_protocol_level > 1:
                    Miscellaneous.protocol(
                        "           Alignment rectangle, set by the user: " + str(
                            y_low_opt) + "<y<" + str(y_high_opt) + ", " + str(
                            x_low_opt) + "<x<" + str(x_high_opt), self.attached_log_file,
                        precede_with_timestamp=False)

        # Align all frames globally relative to the frame with the highest score.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start aligning all frames +++", self.attached_log_file)

        self.my_timer.create_no_check('Global frame alignment')

        # Align all frames in "Surface" mode.
        if self.configuration.align_frames_mode == "Surface":
            # Try the frame alignment using the alignment patch with the highest quality first. If
            # for at least one frame no valid shift can be found, try the next alignment patch. If
            # valid shifts cannot be computed with any patch, abort processing of this job and go to
            # the next one.
            number_patches = len(self.align_frames.alignment_rect_qualities)
            for patch_index in range(number_patches):
                self.align_frames.select_alignment_rect(patch_index)
                try:
                    self.align_frames.align_frames()
                    # Everything is fine, no need to try another stabilization patch.
                    break
                except (NotSupportedError, ArgumentError) as e:
                    self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
                    self.my_timer.stop('Global frame alignment')
                    return
                # For some frames no valid shift could be computed. This would create problems later
                # in the workflow. Therefore, try again with another stabilization patch.
                except InternalError as e:
                    if self.configuration.global_parameters_protocol_level > 0:
                        Miscellaneous.protocol("Warning: No valid shift computed at " + e.message +
                                               ", will try another stabilization patch",
                                               self.attached_log_file)
                    # If there is no more patch available, skip this job.
                    if patch_index == number_patches - 1:
                        if self.configuration.align_frames_search_width < self.configuration.align_frames_max_search_width:
                            self.abort_job_signal.emit(
                                "Error: Frame stabilization failed at " + e.message +
                                ", continuing with next job. "
                                "Try a higher value for parameter 'stabilization search width'.")
                        else:
                            self.abort_job_signal.emit(
                                "Error: Frame stabilization failed at " + e.message +
                                ", continuing with next job. "
                                "Try stabilizing the frames with another program, e.g. PIPP.")
                        self.my_timer.stop('Global frame alignment')
                        return
                    # Continue with the next best stabilization patch.
                    else:
                        y_low_opt, y_high_opt, x_low_opt, x_high_opt = \
                            self.align_frames.alignment_rect_bounds[patch_index + 1]
                        if self.configuration.global_parameters_protocol_level > 0:
                            Miscellaneous.protocol(
                                "           Next alignment rectangle tried: " + str(
                                    y_low_opt) + "<y<" + str(y_high_opt) + ", " + str(
                                    x_low_opt) + "<x<" + str(x_high_opt), self.attached_log_file,
                                precede_with_timestamp=False)

                # Catch all other exceptions. In this case abort the job and try the next one.
                except Exception as e:
                    self.abort_job_signal.emit("Error: " + str(e) + ", continuing with next job")
                    self.my_timer.stop('Global frame alignment')
                    return

        # Align all frames in "Planet" mode.
        else:
            try:
                self.align_frames.align_frames()
            except Error as e:
                self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
                self.my_timer.stop('Global frame alignment')
                return
            except Exception as e:
                self.abort_job_signal.emit(
                    "Error in aligning frames: " + str(e) + ", continuing with next job")
                self.my_timer.stop('Global frame alignment')
                return

        self.my_timer.stop('Global frame alignment')

        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol("           Pixel range common to all frames: " + str(
                self.align_frames.intersection_shape[0][0]) + "<y<" + str(
                self.align_frames.intersection_shape[0][1]) + ", " + str(
                self.align_frames.intersection_shape[1][0]) + "<x<" + str(
                self.align_frames.intersection_shape[1][1]), self.attached_log_file,
                                   precede_with_timestamp=False)

            # Compute and print the maximum shift between two consecutive frames:
            max_shift_y = max([abs(
                self.align_frames.frame_shifts[idx][0] - self.align_frames.frame_shifts[idx - 1][0])
                               for idx in range(1, self.frames.number)])
            max_shift_x = max([abs(
                self.align_frames.frame_shifts[idx][1] - self.align_frames.frame_shifts[idx - 1][1])
                               for idx in range(1, self.frames.number)])
            Miscellaneous.protocol(
                "           Maximal pixel shift between consecutive frames, vertical: " + str(
                    max_shift_y) + ", horizontal: " + str(max_shift_x), self.attached_log_file,
                                   precede_with_timestamp=False)

        # Compute the average frame.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start computing the reference frame +++",
                                   self.attached_log_file)
        self.my_timer.create_no_check('Compute reference frame')
        self.align_frames.average_frame()
        self.my_timer.stop('Compute reference frame')
        if self.configuration.global_parameters_protocol_level > 1:
            if self.configuration.align_frames_fast_changing_object:
                Miscellaneous.protocol(
                    "           The reference frame was computed using the best " + str(
                        self.align_frames.average_frame_number) + " frames within a window of " +
                    str(self.align_frames.average_frame_number *
                        self.configuration.align_frames_best_frames_window_extension) +
                    " frames.\n           Quality loss of reference frame due to time restriction: " +
                    str(self.align_frames.quality_loss_percent) +
                    "%\n           Position of reference frame in video time line: " +
                    str(self.align_frames.cog_mean_frame) + "%",
                    self.attached_log_file, precede_with_timestamp=False)
            else:
                Miscellaneous.protocol(
                    "           The reference frame was computed using the best " + str(
                        self.align_frames.average_frame_number) + " frames.",
                    self.attached_log_file, precede_with_timestamp=False)

        self.work_next_task_signal.emit("Select stack size")

    @QtCore.pyqtSlot(int, int, int, int)
    def execute_set_roi(self, y_min, y_max, x_min, x_max):

        self.set_status_bar_processing_phase("setting the ROI")
        if self.configuration.global_parameters_protocol_level > 0 and (y_min != 0 or y_max != 0):
            Miscellaneous.protocol("+++ Start setting a ROI and computing a new reference frame +++",
                                   self.attached_log_file)
        self.my_timer.create_no_check('Setting ROI and new reference')
        self.align_frames.set_roi(y_min, y_max, x_min, x_max)
        self.my_timer.stop('Setting ROI and new reference')

        if self.configuration.global_parameters_protocol_level > 1 and (y_min != 0 or y_max != 0):
            Miscellaneous.protocol(
                "           ROI, set by the user: " + str(y_min) + "<y<" + str(y_max) + ", " + str(
                    x_min) + "<x<" + str(x_max), self.attached_log_file,
                precede_with_timestamp=False)

        self.work_next_task_signal.emit("Set alignment points")

    @QtCore.pyqtSlot()
    def execute_set_alignment_points(self):

        # If not executing in "automatic" mode, the APs are created on the main_gui thread.
        if self.main_gui.automatic:
            self.set_status_bar_processing_phase("creating alignment points")
            # Initialize the AlignmentPoints object.
            self.my_timer.create_no_check('Initialize alignment point object')
            self.alignment_points = AlignmentPoints(self.configuration, self.frames,
                                                    self.rank_frames, self.align_frames,
                                                    progress_signal=self.work_current_progress_signal)
            self.my_timer.stop('Initialize alignment point object')

            # Create alignment points, and create an image with wll alignment point boxes and
            # patches.
            if self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("+++ Start creating alignment points +++",
                                       self.attached_log_file)
            self.my_timer.create_no_check('Create alignment points')

            # If a ROI is selected, alignment points are created in the ROI window only.
            self.alignment_points.create_ap_grid()

            self.my_timer.stop('Create alignment points')

        self.work_next_task_signal.emit("Compute frame qualities")

    @QtCore.pyqtSlot()
    def execute_compute_frame_qualities(self):

        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol("           Number of alignment points selected: " + str(len(
                self.alignment_points.alignment_points)) + ", aps dropped because too dim: " + str(
                self.alignment_points.alignment_points_dropped_dim) + ", aps dropped because too "
                                                                      "little structure: " + str(
                self.alignment_points.alignment_points_dropped_structure), self.attached_log_file,
                                   precede_with_timestamp=False)

        self.set_status_bar_processing_phase("ranking all frames at all alignment points")
        # For each alignment point rank frames by their quality.
        self.my_timer.create_no_check('Rank frames at alignment points')
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start ranking all frames at all alignment points +++",
                                   self.attached_log_file)
        self.alignment_points.compute_frame_qualities()
        self.my_timer.stop('Rank frames at alignment points')

        self.work_next_task_signal.emit("Stack frames")

    @QtCore.pyqtSlot()
    def execute_stack_frames(self):

        self.set_status_bar_processing_phase("stacking frames")
        # Allocate StackFrames object.
        self.stack_frames = StackFrames(self.configuration, self.frames, self.rank_frames,
                                        self.align_frames, self.alignment_points, self.my_timer,
                                        progress_signal=self.work_current_progress_signal,
                                        debug=self.debug_AP,
                                        create_image_window_signal=self.create_image_window_signal,
                                        update_image_window_signal=self.update_image_window_signal,
                                        terminate_image_window_signal=self.terminate_image_window_signal)

        # Stack all frames.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol(
                "+++ Start stacking " + str(self.alignment_points.stack_size) + " frames +++",
                self.attached_log_file)
        try:
            self.stack_frames.stack_frames()
        except Error as e:
            self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
            return
        except Exception as e:
            self.abort_job_signal.emit(
                "Error in stacking frames: " + str(e) + ", continuing with next job")
            return

        if self.configuration.global_parameters_protocol_level > 1 and len(
            self.alignment_points.alignment_points) > 0:
            Miscellaneous.protocol("\n           Distribution of shifts at alignment points:",
                                   self.attached_log_file, precede_with_timestamp=False)
            Miscellaneous.protocol(self.stack_frames.print_shift_table() + "\n",
                                   self.attached_log_file, precede_with_timestamp=False)

        self.set_status_bar_processing_phase("merging AP patches")
        # Merge the stacked alignment point buffers into a single image.
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start merging all alignment patches and the background +++",
                                   self.attached_log_file)

        try:
            self.stack_frames.merge_alignment_point_buffers()
        except Error as e:
            self.abort_job_signal.emit("Error: " + e.message + ", continuing with next job")
            return
        except Exception as e:
            self.abort_job_signal.emit(
                "Error in merging AP patches: " + str(e) + ", continuing with next job")
            return

        if self.stack_frames.border_y_low or self.stack_frames.border_y_high or self.stack_frames.border_x_low or self.stack_frames.border_x_high:
            if self.configuration.global_parameters_protocol_level > 1:
                Miscellaneous.protocol(
                    "           Stacked image borders cropped to avoid artifacts; top: " + str(
                        self.stack_frames.border_y_low) + ", bottom: " + str(
                        self.stack_frames.border_y_high) + ", left: " + str(
                        self.stack_frames.border_x_low) + ", right: " + str(
                        self.stack_frames.border_x_high) + "\n",
                        self.attached_log_file, precede_with_timestamp=False)

        # If the drizzle factor is 1.5, reduce the pixel resolution of the stacked image buffer
        # to half the size used in stacking.
        if self.configuration.drizzle_factor_is_1_5:
            self.stack_frames.half_stacked_image_buffer_resolution()

        self.work_next_task_signal.emit("Save stacked image")

    @QtCore.pyqtSlot()
    def execute_save_stacked_image(self):

        self.set_status_bar_processing_phase("saving result")

        # Create suffix containing parameter info.
        parameter_suffix = self.compose_suffix()
        if self.job.type == 'video':
            self.stacked_image_name = splitext(self.job.name)[0] + \
                                      self.configuration.stack_frames_suffix + \
                                      parameter_suffix + '.' + \
                                      self.configuration.global_parameters_image_format
        else:  # self.job.type == 'image'
            self.stacked_image_name = self.job.name + self.configuration.stack_frames_suffix + \
                                      parameter_suffix + '.' + \
                                      self.configuration.global_parameters_image_format

        # Save the image as 16bit int (color or mono).
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start saving the stacked image +++", self.attached_log_file)
        self.my_timer.create_no_check('Saving the stacked image')
        Frames.save_image(self.stacked_image_name, self.stack_frames.stacked_image,
                          color=self.frames.color, avoid_overwriting=False,
                          header=self.configuration.global_parameters_version)
        self.my_timer.stop('Saving the stacked image')
        if self.configuration.global_parameters_protocol_level > 1:
            Miscellaneous.protocol(
                "           The stacked image was written to: " + self.stacked_image_name,
                self.attached_log_file, precede_with_timestamp=False)

        # If parameter info is to be included in output file names, compose the new name for
        # the attached log file. The existing file will be renamed after closing.
        if self.attached_log_file:
            self.attached_log_name_new = self.attached_log_name[:self.attached_log_name.index(
                '_stacking-log')] + '_stacking-log' + parameter_suffix + '.txt'

        # If postprocessing is included after stacking, set the stacked image as input.
        if self.configuration.global_parameters_include_postprocessing:
            self.postproc_input_image = self.stack_frames.stacked_image
            self.postproc_input_name = self.stacked_image_name
            self.postprocessed_image_name = PostprocDataObject.set_file_name_processed(
                self.stacked_image_name, self.configuration.postproc_suffix,
                self.configuration.global_parameters_image_format)
            self.work_next_task_signal.emit("Postprocessing")
        else:
            self.work_next_task_signal.emit("Next job")

            # Print timing info for this job.
            self.my_timer.stop('Execution over all')
            if self.configuration.global_parameters_protocol_level > 0:
                self.my_timer.protocol(self.attached_log_file)
            if self.attached_log_file:
                self.attached_log_file.close()
                if self.attached_log_name_new and self.attached_log_name_new != self.attached_log_name:
                    # If a logfile with the new name exists, remove it.
                    try:
                        remove(self.attached_log_name_new)
                    except:
                        pass
                    rename(self.attached_log_name, self.attached_log_name_new)
                    self.attached_log_name = self.attached_log_name_new

    def compose_suffix(self):
        """
        If process parameters are to be included in the stacked output file name, compose the
        suffix from the parameters computed during the workflow.

        :return: Additional suffix string
        """
        string = ''
        if self.configuration.global_parameters_parameters_in_filename:
            if self.configuration.global_parameters_stack_number_frames:
                string += '_f' + str(self.alignment_points.stack_size)
            if self.configuration.global_parameters_stack_percent_frames:
                string += '_p' + str(int(round(100*self.alignment_points.stack_size/self.frames.number)))
            if self.configuration.global_parameters_ap_box_size:
                string += '_b' + str(self.configuration.alignment_points_half_box_width*2)
            if self.configuration.global_parameters_ap_number:
                string += '_ap' + str(len(self.alignment_points.alignment_points))

        return string


    @QtCore.pyqtSlot()
    def execute_postprocess_image(self):

        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("+++ Start postprocessing +++", self.attached_log_file)
        self.my_timer.create_no_check('Conputing image postprocessing')

        # Look up parameters of the last postproc version which was selected in interactive mode.
        version_index = self.configuration.postproc_data_object.version_selected
        postproc_layers = self.configuration.postproc_data_object.versions[version_index].layers
        rgb_automatic = self.configuration.postproc_data_object.versions[
            version_index].rgb_automatic
        rgb_gauss_width = self.configuration.postproc_data_object.versions[
            version_index].rgb_gauss_width
        rgb_resolution_index = self.configuration.postproc_data_object.versions[
            version_index].rgb_resolution_index
        shift_red = self.configuration.postproc_data_object.versions[version_index].shift_red
        shift_blue = self.configuration.postproc_data_object.versions[version_index].shift_blue

        try:
            # Auto-align RGB channels, if requested.
            if rgb_automatic:
                sharpening_input, self.configuration.postproc_data_object.versions[
                    version_index].shift_red, self.configuration.postproc_data_object.versions[
                    version_index].shift_blue = Miscellaneous.auto_rgb_align(
                    self.postproc_input_image, self.configuration.postproc_max_shift,
                    interpolation_factor=[1, 2, 4][rgb_resolution_index], reduce_output=True,
                    blur_strength=rgb_gauss_width)
            elif shift_red != (0., 0.) or shift_blue != (0., 0.):
                # Shift the image with the resolution given by the selected interpolation factor.
                interpolation_factor = [1, 2, 4][rgb_resolution_index]
                sharpening_input = Miscellaneous.shift_colors(self.postproc_input_image,
                                                           shift_red, shift_blue,
                                                           interpolate_input=interpolation_factor,
                                                           reduce_output=interpolation_factor)
            else:
                sharpening_input = self.postproc_input_image

            # Apply all sharpening layers of the postprocessing version selected last time.
            self.postprocessed_image = Miscellaneous.post_process(sharpening_input, postproc_layers)

        except Exception as e:
            self.abort_job_signal.emit(
                "Error in postprocessing: " + str(e) + ", continuing with next job")
            return

        self.my_timer.stop('Conputing image postprocessing')

        self.work_next_task_signal.emit("Save postprocessed image")

    @QtCore.pyqtSlot(object)
    def execute_save_postprocessed_image(self, postprocessed_image):

        # The signal payload is None only if the editor was left with "cancel" in interactive mode.
        # In this case, skip saving the result and proceed with the next job.
        if postprocessed_image is not None:

            # Print postprocessing info if sharpening layers have been applied or RGB alignment was
            # active.
            if self.configuration.global_parameters_protocol_level > 1:
                version_selected = self.configuration.postproc_data_object.version_selected
                postproc_version = self.configuration.postproc_data_object.versions[
                    self.configuration.postproc_data_object.version_selected]
                if version_selected or postproc_version.rgb_automatic:
                    Miscellaneous.print_postproc_parameters(postproc_version,
                                                            self.attached_log_file)
                else:
                    Miscellaneous.protocol(
                        "           The image was not modified in postprocessing.",
                        self.attached_log_file, precede_with_timestamp=False)

            self.set_status_bar_processing_phase("saving result")
            # Save the image as 16bit int (color or mono).
            if self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("+++ Start saving the postprocessed image +++",
                                       self.attached_log_file)
            self.my_timer.create_no_check('Saving the postprocessed image')
            Frames.save_image(self.postprocessed_image_name, postprocessed_image,
                              color=(len(postprocessed_image.shape) == 3), avoid_overwriting=False,
                              header=self.configuration.global_parameters_version)
            self.my_timer.stop('Saving the postprocessed image')
            if self.configuration.global_parameters_protocol_level > 1:
                Miscellaneous.protocol(
                    "           The postprocessed image was written to: " +
                    self.postprocessed_image_name,
                    self.attached_log_file, precede_with_timestamp=False)

        self.work_next_task_signal.emit("Next job")

        # Print timing info for this job.
        self.my_timer.stop('Execution over all')
        if self.configuration.global_parameters_protocol_level > 0:
            self.my_timer.protocol(self.attached_log_file)
        if self.attached_log_file:
            self.attached_log_file.close()
            # If the attached log name was defined for a stacking job, rename it to include
            # parameter information.
            if self.attached_log_name_new and self.attached_log_name_new != self.attached_log_name:
                try:
                    remove(self.attached_log_name_new)
                except:
                    pass
                rename(self.attached_log_name, self.attached_log_name_new)
                self.attached_log_name = self.attached_log_name_new

    def set_status_bar_processing_phase(self, phase):
        """
        Put a text of the form
            "Processing < job name >, < processing step >."
        on the main window status bar.

        :param phase: Processing phase (string)
        :return: -
        """

        self.set_status_bar_signal.emit("Processing " + self.input_name + ", " + phase + ".",
                                        "black")



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from glob import glob
from itertools import chain
from time import time

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from math import ceil
from numpy import float32, zeros, empty, int32, uint8, uint16, clip, histogram
from cv2 import imwrite, moments, threshold, THRESH_BINARY, Laplacian, CV_32F, minMaxLoc

from configuration import Configuration
from exceptions import WrongOrderingError, NotSupportedError, InternalError, ArgumentError, Error
from frames import Frames
from miscellaneous import Miscellaneous
from rank_frames import RankFrames


class AlignFrames(object):
    """
        Based on a list of frames and a (parallel) list of frame quality values, an averaged
        reference frame is created and all frames are aligned with this frame. The alignment is
        performed on a small rectangular area where structure is optimal in both x and y directions.

    """

    def __init__(self, frames, rank_frames, configuration, progress_signal=None):
        """
        Initialize the AlignFrames object with info from the objects "frames" and "rank_frames".

        :param frames: Frames object with all video frames
        :param rank_frames: RankFrames object with global quality ranks (between 0. and 1.,
                            1. being optimal) for all frames
        :param configuration: Configuration object with parameters
        :param progress_signal: Either None (no progress signalling), or a signal with the signature
                                (str, int) with the current activity (str) and the progress in
                                percent (int).
        """

        self.frames = frames
        self.rank_frames = rank_frames
        self.shape = frames.shape
        self.alignment_rect_qualities = None
        self.alignment_rect_bounds = None
        self.frame_shifts = None
        self.intersection_shape = None
        self.intersection_shape_original = None
        self.mean_frame = None
        self.mean_frame_original = None
        self.quality_loss_percent = None
        self.cog_mean_frame = None
        self.configuration = configuration
        self.progress_signal = progress_signal
        self.signal_step_size = max(int(self.frames.number / 10), 1)
        self.x_low_opt = self.x_high_opt = self.y_low_opt = self.y_high_opt = None
        self.dy = self.dx = None
        self.y_low_opt = self.y_high_opt = self.x_low_opt = self.x_high_opt = None
        self.dy_original = self.dx_original = None
        self.ROI_set = False
        self.dev_table = empty((2 * self.configuration.align_frames_search_width,
                               2 * self.configuration.align_frames_search_width), dtype=float32)

    def compute_alignment_rect(self, scale_factor):
        """
        Using the frame with the highest rank (sharpest image), compute the rectangular patch
        where structure is best in both x and y directions. The size of the patch is the size of the
        frame divided by "scale_factor" in both coordinate directions.

        :param scale_factor: Ratio of the size of the frame and the alignment patch in both
                             coordinate directions
        :return: A four tuple (y_low, y_high, x_low, x_high) with pixel coordinates of the
                 optimal patch
        """

        dim_y, dim_x = self.shape[0:2]

        # Compute the extensions of the alignment rectangle in y and x directions. Take into account
        # a border and the search radius for frame shifts.
        border_width = self.configuration.align_frames_border_width + \
                       self.configuration.align_frames_search_width
        rect_y = int((dim_y - 2 * border_width) / scale_factor)
        rect_y_2 = int(rect_y / 2)
        rect_x = int((dim_x - 2 * border_width) / scale_factor)
        rect_x_2 = int(rect_x / 2)

        # Initialize lists which for each location store the quality measure and the index bounds.
        self.alignment_rect_qualities = []
        self.alignment_rect_bounds = []

        # Compute for all locations in the frame the "quality measure" and store the location
        # and quality measure in above lists.
        best_frame_mono_blurred = self.frames.frames_mono_blurred(self.rank_frames.frame_ranks_max_index)
        x_low = border_width
        x_high = x_low + rect_x
        while x_high <= dim_x - border_width:
            y_low = border_width
            y_high = y_low + rect_y
            while y_high <= dim_y - border_width:
                # new_quality = Miscellaneous.local_contrast(
                #     self.frames_mono_blurred[self.frame_ranks_max_index][y_low:y_high,
                #     x_low:x_high], self.configuration.quality_area_pixel_stride)
                self.alignment_rect_qualities.append(
                    Miscellaneous.quality_measure_threshold_weighted(
                    best_frame_mono_blurred[y_low:y_high, x_low:x_high],
                    stride=self.configuration.align_frames_rectangle_stride,
                    black_threshold=self.configuration.align_frames_rectangle_black_threshold,
                    min_fraction=self.configuration.align_frames_rectangle_min_fraction))
                self.alignment_rect_bounds.append((y_low, y_high, x_low, x_high))

                y_low += rect_y_2
                y_high += rect_y_2
            x_low += rect_x_2
            x_high += rect_x_2

        try:
            # Sort lists by quality.
            arq, arb = zip(*sorted(zip(self.alignment_rect_qualities, self.alignment_rect_bounds),
                                   reverse=True))
            self.alignment_rect_qualities = list(arq)
            self.alignment_rect_bounds = list(arb)

            # Set the optimal coordinates and return them as a tuple.
            (self.y_low_opt, self.y_high_opt, self.x_low_opt, self.x_high_opt) = \
                self.alignment_rect_bounds[0]
            return self.alignment_rect_bounds[0]
        except:
            raise ArgumentError("The frame is too small for alignment parameters chosen")

    def select_alignment_rect(self, index):
        """
        Select an alignment patch from the list computed in "compute_aligment_rect" to be used
        for frame aligment.

        :param index: index of the alignment patch in the list in decreasing quality order.
        :return: True, if successful. Otherwise return False.
        """

        if index < 0 or index > len(self.alignment_rect_bounds):
            return False
        else:
            self.y_low_opt, self.y_high_opt, self.x_low_opt, self.x_high_opt = \
                self.alignment_rect_bounds[index]
            return True

    def set_alignment_rect(self, y_low_opt, y_high_opt, x_low_opt, x_high_opt):
        """
        As an alternative to "select_alignment_rect", the rectangular patch can be set explicitly

        :param y_low_opt: Lower y pixel coordinate of patch
        :param y_high_opt: Upper y pixel coordinate of patch
        :param x_low_opt: Lower x pixel coordinate of patch
        :param x_high_opt: Upper x pixel coordinate of patch
        :return: -
        """

        self.y_low_opt = y_low_opt
        self.y_high_opt = y_high_opt
        self.x_low_opt = x_low_opt
        self.x_high_opt = x_high_opt

        # If the user has set the patch manually, set the lists to this single element.
        self.alignment_rect_qualities = [1.]
        self.alignment_rect_bounds = [(self.y_low_opt, self.y_high_opt, self.x_low_opt,
                                       self.x_high_opt)]

    def align_frames(self):
        """
        Compute the displacement of all frames relative to the sharpest frame using the alignment
        rectangle.

        :return: -
        """

        if self.configuration.align_frames_mode == "Surface":
            # For "Surface" mode the alignment rectangle has to be selected first.
            if self.x_low_opt is None:
                raise WrongOrderingError(
                    "Method 'align_frames' is called before 'select_alignment_rect'")

            # From the sharpest frame cut out the alignment rectangle. The shifts of all other frames
            # will be computed relativ to this patch.
            if self.configuration.align_frames_method == "MultiLevelCorrelation":
                # MultiLevelCorrelation uses two reference windows with different resolution. Also,
                # please note that the data type is float32 in this case.
                reference_frame = self.frames.frames_mono_blurred(
                    self.rank_frames.frame_ranks_max_index).astype(float32)
                self.reference_window = reference_frame[self.y_low_opt:self.y_high_opt,
                                        self.x_low_opt:self.x_high_opt]
                # For the first phase a box with half the resolution is constructed.
                self.reference_window_first_phase = self.reference_window[::2, ::2]
            else:
                # For all other methods, the reference window is of type int32.
                reference_frame = self.frames.frames_mono_blurred(
                    self.rank_frames.frame_ranks_max_index).astype(int32)
                self.reference_window = reference_frame[
                                        self.y_low_opt:self.y_high_opt,
                                        self.x_low_opt:self.x_high_opt]

            self.reference_window_shape = self.reference_window.shape

        elif self.configuration.align_frames_mode == "Planet":
            # For "Planetary" mode compute the center of gravity for the reference image.
            cog_reference_y, cog_reference_x = AlignFrames.center_of_gravity(
                self.frames.frames_mono_blurred(self.rank_frames.frame_ranks_max_index))

        else:
            raise NotSupportedError(
                "Frame alignment mode '" + self.configuration.align_frames_mode +
                "' not supported")

        # Initialize a list which for each frame contains the shifts in y and x directions.
        self.frame_shifts = [None] * self.frames.number

        # Initialize a counter of processed frames for progress bar signalling. It is set to one
        # because in the loop below the optimal frame is not counted.
        number_processed = 1

        # Loop over all frames. Begin with the sharpest (reference) frame
        for idx in chain(reversed(range(self.rank_frames.frame_ranks_max_index + 1)),
                         range(self.rank_frames.frame_ranks_max_index, self.frames.number)):

            if idx == self.rank_frames.frame_ranks_max_index:
                # For the sharpest frame the displacement is 0 because it is used as the reference.
                self.frame_shifts[idx] = [0, 0]
                # Initialize two variables which keep the shift values of the previous step as
                # the starting point for the next step. This reduces the search radius if frames are
                # drifting.
                dy_min_cum = dx_min_cum = 0

            # For all other frames: Compute the global shift, using the "blurred" monochrome image.
            else:
                # After every "signal_step_size"th frame, send a progress signal to the main GUI.
                if self.progress_signal is not None and number_processed % self.signal_step_size == 1:
                    self.progress_signal.emit("Align all frames",
                                        int(round(10*number_processed / self.frames.number) * 10))

                frame = self.frames.frames_mono_blurred(idx)

                # In Planetary mode the shift of the "center of gravity" of the image is computed.
                # This algorithm cannot fail.
                if self.configuration.align_frames_mode == "Planet":

                    cog_frame = AlignFrames.center_of_gravity(frame)
                    self.frame_shifts[idx] = [cog_reference_y - cog_frame[0],
                                              cog_reference_x - cog_frame[1]]

                # In Surface mode various methods can be used to measure the shift from one frame
                # to the next. The method "Translation" is special: Using phase correlation it is
                # the only method not based on a local search algorithm. It is treated differently
                # here because it does not require a re-shifting of the alignment patch.
                elif self.configuration.align_frames_method == "Translation":
                    # The shift is computed with cross-correlation. Cut out the alignment patch and
                    # compute its translation relative to the reference.
                    frame_window = self.frames.frames_mono_blurred(idx)[
                                   self.y_low_opt:self.y_high_opt, self.x_low_opt:self.x_high_opt]
                    self.frame_shifts[idx] = Miscellaneous.translation(self.reference_window,
                                                                       frame_window,
                                                                       self.reference_window_shape)

                # Now treat all "Surface" mode cases using local search algorithms. In each case
                # the result is the shift vector [dy_min, dx_min]. The search can fail (if within
                # the search radius no optimum is found). If that happens for at least one frame,
                # an exception is raised. The workflow thread then tries again using another
                # alignment patch.
                else:

                    if self.configuration.align_frames_method == "MultiLevelCorrelation":
                        # The shift is computed in two phases: First on a coarse pixel grid,
                        # and then on the original grid in a small neighborhood around the optimum
                        # found in the first phase.
                        shift_y_local_first_phase, shift_x_local_first_phase, \
                        success_first_phase, shift_y_local_second_phase, \
                        shift_x_local_second_phase, success_second_phase = \
                            Miscellaneous.multilevel_correlation(
                            self.reference_window_first_phase, frame,
                            self.configuration.frames_gauss_width,
                            self.reference_window, self.y_low_opt - dy_min_cum,
                                                          self.y_high_opt - dy_min_cum,
                                                          self.x_low_opt - dx_min_cum,
                                                          self.x_high_opt - dx_min_cum,
                            self.configuration.align_frames_search_width,
                            weight_matrix_first_phase=None)

                        success = success_first_phase and success_second_phase
                        if success:
                            [dy_min, dx_min] = [
                                shift_y_local_first_phase + shift_y_local_second_phase,
                                shift_x_local_first_phase + shift_x_local_second_phase]


                    elif self.configuration.align_frames_method == "RadialSearch":
                        # Spiral out from the shift position of the previous frame and search for the
                        # local optimum.
                        [dy_min, dx_min], dev_r = Miscellaneous.search_local_match(
                            self.reference_window, frame, self.y_low_opt - dy_min_cum,
                                                          self.y_high_opt - dy_min_cum,
                                                          self.x_low_opt - dx_min_cum,
                                                          self.x_high_opt - dx_min_cum,
                            self.configuration.align_frames_search_width,
                            self.configuration.align_frames_sampling_stride, sub_pixel=False)

                        # The search was not successful if a zero shift was reported after more
                        # than two search cycles.
                        success = len(dev_r) <= 2 or dy_min != 0 or dx_min != 0

                    elif self.configuration.align_frames_method == "SteepestDescent":
                        # Spiral out from the shift position of the previous frame and search for the
                        # local optimum.
                        [dy_min, dx_min], dev_r = Miscellaneous.search_local_match_gradient(
                            self.reference_window, frame, self.y_low_opt - dy_min_cum,
                                                          self.y_high_opt - dy_min_cum,
                                                          self.x_low_opt - dx_min_cum,
                                                          self.x_high_opt - dx_min_cum,
                            self.configuration.align_frames_search_width,
                            self.configuration.align_frames_sampling_stride, self.dev_table)

                        # The search was not successful if a zero shift was reported after more
                        # than two search cycles.
                        success = len(dev_r) <= 2 or dy_min != 0 or dx_min != 0

                    else:
                        raise NotSupportedError(
                            "Frame alignment method " + configuration.align_frames_method +
                            " not supported")

                    # If the local search was unsuccessful, quit the frame loop with an error.
                    if not success:
                        raise InternalError("frame " + str(idx))

                    # Update the cumulative shift values to be used as starting point for the
                    # next frame.
                    dy_min_cum += dy_min
                    dx_min_cum += dx_min
                    self.frame_shifts[idx] = [dy_min_cum, dx_min_cum]

                    # If the alignment window gets too close to a frame edge, move it away from
                    # that edge by half the border width. First check if the reference window still
                    # fits into the shifted frame.
                    if self.shape[0] - abs(
                            dy_min_cum) - 2 * self.configuration.align_frames_search_width - \
                            self.configuration.align_frames_border_width < \
                            self.reference_window_shape[0] or self.shape[1] - abs(
                            dx_min_cum) - 2 * self.configuration.align_frames_search_width - \
                            self.configuration.align_frames_border_width < \
                            self.reference_window_shape[1]:
                        raise ArgumentError("Frame stabilization window does not fit into"
                                            " intersection")

                    new_reference_window = False
                    # Start with the lower y edge.
                    while self.y_low_opt - dy_min_cum < \
                            self.configuration.align_frames_search_width + \
                            self.configuration.align_frames_border_width / 2:
                        self.y_low_opt += ceil(self.configuration.align_frames_border_width / 2.)
                        self.y_high_opt += ceil(self.configuration.align_frames_border_width / 2.)
                        new_reference_window = True
                    # Now the upper y edge.
                    while self.y_high_opt - dy_min_cum > self.shape[
                        0] - self.configuration.align_frames_search_width - \
                            self.configuration.align_frames_border_width / 2:
                        self.y_low_opt -= ceil(self.configuration.align_frames_border_width / 2.)
                        self.y_high_opt -= ceil(self.configuration.align_frames_border_width / 2.)
                        new_reference_window = True
                    # Now the lower x edge.
                    while self.x_low_opt - dx_min_cum < \
                            self.configuration.align_frames_search_width + \
                            self.configuration.align_frames_border_width / 2:
                        self.x_low_opt += ceil(self.configuration.align_frames_border_width / 2.)
                        self.x_high_opt += ceil(self.configuration.align_frames_border_width / 2.)
                        new_reference_window = True
                    # Now the upper x edge.
                    while self.x_high_opt - dx_min_cum > self.shape[
                        1] - self.configuration.align_frames_search_width - \
                            self.configuration.align_frames_border_width / 2:
                        self.x_low_opt -= ceil(self.configuration.align_frames_border_width / 2.)
                        self.x_high_opt -= ceil(self.configuration.align_frames_border_width / 2.)
                        new_reference_window = True

                    # If the window was moved, update the "reference window(s)".
                    if new_reference_window:
                        if self.configuration.align_frames_method == "MultiLevelCorrelation":
                            self.reference_window = reference_frame[self.y_low_opt:self.y_high_opt,
                                                    self.x_low_opt:self.x_high_opt]
                            # For the first phase a box with half the resolution is constructed.
                            self.reference_window_first_phase = self.reference_window[::2, ::2]
                        else:
                            self.reference_window = reference_frame[self.y_low_opt:self.y_high_opt,
                                                    self.x_low_opt:self.x_high_opt]

                # This frame is processed, go to next one.
                number_processed += 1

        if self.progress_signal is not None:
            self.progress_signal.emit("Align all frames", 100)

        # Compute the shape of the area contained in all frames in the form [[y_low, y_high],
        # [x_low, x_high]]
        self.intersection_shape = [[max(b[0] for b in self.frame_shifts),
                                    min(b[0] for b in self.frame_shifts) + self.shape[0]],
                                   [max(b[1] for b in self.frame_shifts),
                                    min(b[1] for b in self.frame_shifts) + self.shape[1]]]

    @staticmethod
    def center_of_gravity(frame):
        """
        Comppute (y, x) pixel coordinates of the center of gravity for a given monochrome frame.
        Raise an error if the computed cog is outside the frame index bounds.

        :param frame: Monochrome frame (2D numpy array)
        :return: Integer pixel coordinates (center_y, center_x) of center of gravity
        """

        # The following is the old algorithm (up to Version 0.8.5). It does not work well for
        # planets on a bright sky background.
        #
        # Convert the grayscale image to binary image, where all pixels
        # brighter than half the maximum image brightness are set to 1,
        # and all others are set to 0.
        # thresh = threshold(frame, frame.max()/2, 1, THRESH_BINARY)[1]

        # This new code sets the threshold between the minimal brightness (background) and
        # the maximal brightness (object). The hope is that this way the threshold is far away from
        # background noise. Also, no binary image is created, but brightness variations are allowed
        # to influence the center of gravity. This gives brighter parts of the image more weight,
        # which results in a slightly better precision.
        minVal, maxVal, minLoc, maxLoc = minMaxLoc(frame)
        brightness_threshold = int((minVal+maxVal)/2)
        thresh = clip(frame, brightness_threshold, None)[:,:]-brightness_threshold

        # Calculate moments of binary image
        M = moments(thresh)

        # Calculate coordinates for center of gravity and round pixel
        # coordinates to the nearest integers.
        cog_x = round(M["m10"] / M["m00"])
        cog_y = round(M["m01"] / M["m00"])

        # If the computed center of gravity is outside the frame bounds, raise an error (should be
        # impossible).
        if not 0 < cog_y < frame.shape[0] or not 0 < cog_x < frame.shape[1]:
            raise InternalError(
                "Center of gravity coordinates [" + str(cog_y) + ", " + str(
                    cog_x) + "] of reference frame are out of bounds")

        return cog_y, cog_x

    def average_frame(self, average_frame_number=None, color=False):
        """
        Compute an averaged frame from the best (monochrome) frames.

        :param average_frame_number: Number of best frames to be averaged. If None, the number is
                                     computed from the configuration parameter
                                      "align_frames_average_frame_percent"
        :param color: If True, compute an average of the original (color) images. Otherwise use the
                      monochrome frame versions.
        :return: The averaged frame
        """

        if self.intersection_shape is None:
            raise WrongOrderingError("Method 'average_frames' is called before 'align_frames'")

        # Compute global offsets of current frame relative to intersection frame. Start with
        # Initializing lists which for each frame give the dy and dx displacements between the
        # reference frame and current frame.
        self.dy = [self.intersection_shape[0][0] - self.frame_shifts[idx][0] for idx in range(self.frames.number)]
        self.dx = [self.intersection_shape[1][0] - self.frame_shifts[idx][1] for idx in range(self.frames.number)]

        # If the number of frames is not specified explicitly, compute it from configuration.
        if average_frame_number is not None:
            self.average_frame_number = average_frame_number
        else:
            self.average_frame_number = max(
                ceil(self.frames.number * self.configuration.align_frames_average_frame_percent / 100.), 1)

        # If the object is changing fast (Jupiter, Sun), restrict interval from which frames are
        # taken for computing the reference frame.
        if self.configuration.align_frames_fast_changing_object:
            window_size = min(
                self.average_frame_number *
                self.configuration.align_frames_best_frames_window_extension,
                self.frames.number)
            average_frame_indices, self.quality_loss_percent, self.cog_mean_frame = \
                self.rank_frames.find_best_frames(self.average_frame_number, window_size)
        else:
            average_frame_indices = self.rank_frames.quality_sorted_indices[:self.average_frame_number]

        # Create an empty numpy buffer. The first and second dimensions are the y and x
        # coordinates. For color frames add a third dimension. Add all frames to the buffer.
        if color:
            self.mean_frame = zeros([self.intersection_shape[0][1] - self.intersection_shape[0][0],
                 self.intersection_shape[1][1] - self.intersection_shape[1][0], 3], dtype=float32)
            for frame_index in average_frame_indices:
                shift = self.frame_shifts[frame_index]
                self.mean_frame += self.frames.frames(frame_index) \
                    [self.intersection_shape[0][0] - shift[0]:
                    self.intersection_shape[0][1] - shift[0],
                    self.intersection_shape[1][0] - shift[1]:
                    self.intersection_shape[1][1] - shift[1], :]
        else:
            self.mean_frame = zeros([self.intersection_shape[0][1] - self.intersection_shape[0][0],
                                     self.intersection_shape[1][1] - self.intersection_shape[1][0]],
                                     dtype=float32)
            for frame_index in average_frame_indices:
                shift = self.frame_shifts[frame_index]
                self.mean_frame += self.frames.frames_mono(frame_index) \
                    [self.intersection_shape[0][0] - shift[0]:
                    self.intersection_shape[0][1] - shift[0],
                    self.intersection_shape[1][0] - shift[1]:
                    self.intersection_shape[1][1] - shift[1]]

        # Compute the mean frame by dividing by the number of frames, and convert values to 16bit.
        if self.frames.dt0 == uint8:
            scaling = 256. / self.average_frame_number
        elif self.frames.dt0 == uint16:
            scaling = 1. / self.average_frame_number
        else:
            raise NotSupportedError("Attempt to compute the reference frame from images with type"
                                    " neither uint8 nor uint16")

        self.mean_frame = (self.mean_frame*scaling).astype(int32)

        return self.mean_frame

    def set_roi(self, y_min, y_max, x_min, x_max):
        """
        Make the stacking region snmaller than the intersection size. Be careful: The pixel
        indices in this method refer to the shape of the intersection of all frames, i.e., the
        shape of the full mean frame. In general, the original frames are somewhat larger.

        If all four index bounds are zero, set the ROI to the full frame.

        :param y_min: Lower y pixel bound
        :param y_max: Upper y pixel bound
        :param x_min: Lower x pixel bound
        :param x_max: Upper x pixel bound
        :return: The new averaged frame, restricted to the ROI
        """

        if self.intersection_shape is None:
            raise WrongOrderingError("Method 'set_roi' is called before 'align_frames'")

        # On the first call, keep a copy of the full mean frame and original intersection shape.
        if not self.ROI_set:
            self.mean_frame_original = self.mean_frame.copy()
            self.intersection_shape_original = self.intersection_shape.copy()

        if y_min == 0 and y_max == 0 and x_min == 0 and x_max == 0:
            y_min = 0
            y_max = self.intersection_shape_original[0][1] - \
                self.intersection_shape_original[0][0]
            x_min = 0
            x_max = self.intersection_shape_original[1][1] - \
                self.intersection_shape_original[1][0]
        elif y_min < 0 or y_max > self.intersection_shape_original[0][1] - \
                self.intersection_shape_original[0][0] or \
                x_min < 0 or x_max > self.intersection_shape_original[1][1] - \
                self.intersection_shape_original[1][0] or \
                y_min >= y_max or x_min >= x_max:
            raise ArgumentError("Invalid ROI index bounds specified")

        # Reduce the intersection shape and mean frame to the ROI.
        self.intersection_shape = [[y_min+self.intersection_shape_original[0][0],
                                    y_max+self.intersection_shape_original[0][0]],
                                   [x_min+self.intersection_shape_original[1][0],
                                    x_max+self.intersection_shape_original[1][0]]]

        # Re-compute global offsets of current frame relative to reference frame.
        self.dy = [self.intersection_shape[0][0] - self.frame_shifts[idx][0] for idx in range(self.frames.number)]
        self.dx = [self.intersection_shape[1][0] - self.frame_shifts[idx][1] for idx in range(self.frames.number)]

        self.ROI_set = True

        self.mean_frame = self.mean_frame_original[y_min:y_max, x_min:x_max]

        return self.mean_frame

    def reset_roi(self):
        """
        After a ROI has been set, reset the ROI to the full frame. Restore the mean frame and
        the intersection shape to their original values. If no ROI has been set, do nothing.

        :return: -
        """

        if self.ROI_set:
            self.mean_frame = self.mean_frame_original
            self.intersection_shape = self.intersection_shape_original

    def write_stabilized_video(self, name, fps, stabilized=True):
        """
        Write out a stabilized videos. For all frames the part common to all frames is extracted
        and written into a video file.

        :param name: File name of the video output
        :param fps: Frames per second of video
        :param stabilized: if False, switch off image stabilization. Write original frames.
        :return: -
        """

        # Initialize lists of stabilized frames and index strings.
        frames_mono_stabilized = []
        frame_indices = []

        if stabilized:
            # For each frame: cut out the shifted window with the intersection of all frames.
            # Append it to the list, and add its index to the list of index strings.
            for idx in range(self.frames.number):
                frame_mono = self.frames.frames_mono(idx)
                frames_mono_stabilized.append(frame_mono[
                                              self.intersection_shape[0][0] -
                                              self.frame_shifts[idx][0]:
                                              self.intersection_shape[0][1] -
                                              self.frame_shifts[idx][0],
                                              self.intersection_shape[1][0] -
                                              self.frame_shifts[idx][1]:
                                              self.intersection_shape[1][1] -
                                              self.frame_shifts[idx][1]])
                frame_indices.append(str(idx))
            Miscellaneous.write_video(name, frames_mono_stabilized, frame_indices, 5)
        else:
            # Write the original frames (not stabilized) with index number insertions.
            frames_mono = []
            for idx in range(self.frames.number):
                frame_indices.append(str(idx))
                frames_mono.append(self.frames.frames_mono(idx))
            Miscellaneous.write_video(name, frames_mono, frame_indices, 5)


if __name__ == "__main__":
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        names = glob(
            'Images/2012*.tif')
        # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')
        # names = glob.glob('Images/Example-3*.jpg')
    else:
        # file = 'Moon_Tile-013_205538_short'
        # file = 'another_short_video'
        # file = 'Moon_Tile-024_043939'
        # file = 'Moon_Tile-013_205538'
        file = "Venus_06_01"
        names = 'Videos/' + file + '.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        # In creating the Frames object the images are read from the specified file(s).
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + str(e))
        exit()

    # Rank the frames by their overall local contrast.
    rank_frames = RankFrames(frames, configuration)
    rank_frames.frame_score()

    print("Best frame index: " + str(rank_frames.frame_ranks_max_index))
    output_file = 'Images/' + file + '.jpg'
    imwrite(output_file, frames.frames_mono(rank_frames.frame_ranks_max_index))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)

        # Alternative: Set the alignment rectangle by hand.
        # (align_frames.x_low_opt, align_frames.x_high_opt, align_frames.y_low_opt,
        #  align_frames.y_high_opt) = (x_low_opt, x_high_opt, y_low_opt, y_high_opt) = (
        # 650, 950, 550, 750)

        print("optimal alignment rectangle, x_low: " + str(x_low_opt) + ", x_high: " + str(
            x_high_opt) + ", y_low: " + str(y_low_opt) + ", y_high: " + str(y_high_opt))
        frame = frames.frames_mono_blurred(rank_frames.frame_ranks_max_index).copy()
        frame[y_low_opt, x_low_opt:x_high_opt] = frame[y_high_opt - 1, x_low_opt:x_high_opt] = 255
        frame[y_low_opt:y_high_opt, x_low_opt] = frame[y_low_opt:y_high_opt, x_high_opt - 1] = 255
        plt.imshow(frame, cmap='Greys_r')
        plt.show()

    # Align all frames globally relative to the frame with the highest score.
    try:
        align_frames.align_frames()
    except NotSupportedError as e:
        print("Error: " + e.message)
        exit()
    except InternalError as e:
        print("Warning: " + e.message)

    print("Frame shifts: " + str(align_frames.frame_shifts))
    print("Intersection: " + str(align_frames.intersection_shape))

    # Compute the reference frame by averaging the best frames.
    average = align_frames.average_frame()
    plt.imshow(average, cmap='Greys_r')
    plt.show()

    # Write video with stabilized frames, annotated with their frame indices.
    stabilized = True
    if stabilized:
        name = 'Videos/stabilized_video_with_frame_numbers.avi'
    else:
        name = 'Videos/not_stabilized_video_with_frame_numbers.avi'
    align_frames.write_stabilized_video(name, 5, stabilized=stabilized)

    # print ("Setting ROI and computing new average frame")
    # average = align_frames.set_roi(300, 600, 500, 1000)
    # plt.imshow(average, cmap='Greys_r')
    # plt.show()




# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from time import time, sleep
from collections import OrderedDict

from exceptions import ArgumentError
from miscellaneous import Miscellaneous


class timer(object):
    """
    Measure execution times of code sections. Several timers can be used at the same time. Timers
    can be created, incremented, and deleted.
    """

    def __init__(self):
        """
        Initialize the timer object.
        """

        self.counters = OrderedDict()

    def create(self, name):
        """
        Create a named timer, and start it.

        :param name: Name of the timer
        :return: -
        """

        if name in self.counters.keys():
            raise ArgumentError("Attempt to initialize timer with existing name")
        else:
            self.counters[name] = [0., time()]

    def create_no_check(self, name):
        """
        If a named timer does not exist, create one. If it exists, reset and start it.

        :param name: Name of the timer
        :return: -
        """

        self.counters[name] = [0., time()]

    def delete(self, name):
        """
        Delete a named timer.

        :param name: Name of the timer
        :return: -
        """

        if name not in self.counters.keys():
            raise ArgumentError("Attempt to delete timer with undefined name")
        else:
            del self.counters[name]

    def exists(self, name):
        """
        Find out if a named timer with a given name already exists.

        :param name: Name of the timer
        :return: True, if the timer exists; otherwise False.
        """

        return name in self.counters.keys()

    def start(self, name):
        """
        Start a timer.

        :param name: Name of the timer
        :return: -
        """

        if name not in self.counters.keys():
            raise ArgumentError("Attempt to start timer with undefined name")
        else:
            self.counters[name][1] = time()

    def stop(self, name):
        """
        Stop a timer.

        :param name: Name of the timer
        :return: Elapsed time accumulated during start/stop intervals of this timer so far.
        """

        if name not in self.counters.keys():
            raise ArgumentError("Attempt to stop timer with undefined name")
        else:
            self.counters[name][0] += time() - self.counters[name][1]
            return self.counters[name][0]

    def read(self, name):
        """
        Read out a timer.

        :param name: Name of the timer
        :return: Elapsed time accumulated during start/stop intervals of this timer so far.
        """

        if name not in self.counters.keys():
            raise ArgumentError("Attempt to read out timer with undefined name")
        else:
            return self.counters[name][0]

    def reset(self, name):
        """
        Reset a timer to zero without deleting it.

        :param name: Name of the timer.
        :return: -
        """

        if name not in self.counters.keys():
            raise ArgumentError("Attempt to reset timer with undefined name")
        else:
            self.counters[name][0] = 0.

    def print(self):
        print("--------------------------------------------------\nStatus of time counters:")
        for name in self.counters.keys():

            print("{0:40} {1:8.3f}".format(name, self.counters[name][0]))
        print("--------------------------------------------------")

    def protocol(self, logfile):
        Miscellaneous.protocol("", logfile, precede_with_timestamp=False)
        Miscellaneous.protocol("           --------------------------------------------------\n"
                               "           Status of time counters:", logfile,
                               precede_with_timestamp=False)
        for name in self.counters.keys():
            Miscellaneous.protocol(
                "           {0:40} {1:8.3f}".format(name, self.counters[name][0]), logfile,
                precede_with_timestamp=False)
        Miscellaneous.protocol("           --------------------------------------------------\n",
                               logfile, precede_with_timestamp=False)


if __name__ == "__main__":
    my_timer = timer()
    my_timer.create('Second')
    my_timer.create('First')
    try:
        my_timer.create('First')
    except ArgumentError as e:
        print(e.message)

    my_timer.start('First')
    sleep(0.1)
    print("First timer: " + str(my_timer.stop('First')))

    my_timer.start('First')
    sleep(0.5)
    print("First timer: " + str(my_timer.stop('First')))

    start = time()
    sleep(0.1)
    print("Elapsed time (without timer): " + str(time() - start))

    my_timer.start('Second')
    sleep(0.1)
    print("Second timer: " + str(my_timer.stop('Second')))

    my_timer.reset('First')
    my_timer.start('First')
    sleep(0.5)
    print("First timer: " + str(my_timer.stop('First')))

    my_timer.print()

    my_timer.delete('First')
    try:
        my_timer.read('First')
    except ArgumentError as e:
        print(e.message)



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from glob import glob
from statistics import mean
from time import time

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from cv2 import meanStdDev
from numpy import array, full

from configuration import Configuration
from exceptions import ArgumentError, NotSupportedError, Error
from frames import Frames
from miscellaneous import Miscellaneous


class RankFrames(object):
    """
        Rank frames according to their overall sharpness. Experiments with different algorithms
        have been made. The classical "Sobel" algorithm is good but slow. An alternative is
        implemented in method "local_contrast" in module "miscellaneous".

    """

    def __init__(self, frames, configuration, progress_signal=None):
        """
        Initialize the object and instance variables.

        :param frames: Frames object with all video frames
        :param configuration: Configuration object with parameters
        :param progress_signal: Either None (no progress signalling), or a signal with the signature
                                (str, int) with the current activity (str) and the progress in
                                percent (int).
        """


        self.shape = frames.shape
        self.configuration = configuration
        self.frames = frames

        self.number_original = frames.number
        self.frame_ranks_original = []
        self.quality_sorted_indices_original = None
        self.rank_indices_original = None
        self.frame_ranks_max_index_original = None
        self.frame_ranks_max_value_original = None

        self.number = None
        self.frame_ranks = None
        self.quality_sorted_indices = None
        self.rank_indices = None
        self.frame_ranks_max_index = None
        self.frame_ranks_max_value = None

        self.progress_signal = progress_signal
        self.signal_step_size = max(int(self.number_original / 10), 1)

    def frame_score(self):
        """
        Compute the frame quality values and normalize them such that the best value is 1.

        :return: -
        """

        if self.configuration.rank_frames_method == "xy gradient":
            method = Miscellaneous.local_contrast
        elif self.configuration.rank_frames_method == "Laplace":
            method = Miscellaneous.local_contrast_laplace
        elif self.configuration.rank_frames_method == "Sobel":
            method = Miscellaneous.local_contrast_sobel
        else:
            raise NotSupportedError("Ranking method " + self.configuration.rank_frames_method +
                                    " not supported")

        # Reset frames index translation, if active.
        if self.frames.index_translation_active:
            self.frames.reset_index_translation()

        # For all frames compute the quality with the selected method.
        if method != Miscellaneous.local_contrast_laplace:
            for frame_index in range(self.number_original):
                frame = self.frames.frames_mono_blurred(frame_index)
                if self.progress_signal is not None and frame_index % self.signal_step_size == 1:
                    self.progress_signal.emit("Rank all frames",
                                              int(round(10*frame_index / self.number_original) * 10))
                if self.configuration.frames_normalization:
                    self.frame_ranks_original.append(
                        method(frame, self.configuration.rank_frames_pixel_stride) /
                        self.frames.average_brightness(frame_index))
                else:
                    self.frame_ranks_original.append(
                        method(frame, self.configuration.rank_frames_pixel_stride))
        else:
            for frame_index in range(self.number_original):
                frame = self.frames.frames_mono_blurred_laplacian(frame_index)
                # self.frame_ranks.append(mean((frame - frame.mean())**2))
                if self.progress_signal is not None and frame_index % self.signal_step_size == 1:
                    self.progress_signal.emit("Rank all frames",
                                              int(round(10*frame_index / self.number_original) * 10))
                if self.configuration.frames_normalization:
                    self.frame_ranks_original.append(meanStdDev(frame)[1][0][0] /
                        self.frames.average_brightness(frame_index))
                else:
                    self.frame_ranks_original.append(meanStdDev(frame)[1][0][0])

        # Sort the frame indices in descending order of quality.
        self.quality_sorted_indices_original = sorted(range(self.number_original),
                                             key=self.frame_ranks_original.__getitem__, reverse=True)

        # Compute the inverse index list: For each frame the rank_index is the corresponding index
        # in the sorted frame_ranks list.
        self.rank_indices_original = [self.quality_sorted_indices_original.index(index) for index in
                             range(self.number_original)]

        if self.progress_signal is not None:
            self.progress_signal.emit("Rank all frames", 100)

        # Set the index of the best frame, and normalize all quality values.
        self.frame_ranks_max_index_original = self.quality_sorted_indices_original[0]
        self.frame_ranks_max_value_original = self.frame_ranks_original[self.frame_ranks_max_index_original]
        self.frame_ranks_original /= self.frame_ranks_max_value_original

        # Keep the original ranking data and prepare for index translation. The translation can be
        # reset later, and the original ranking be re-established.
        self.number = self.number_original
        self.frame_ranks = self.frame_ranks_original
        self.quality_sorted_indices = self.quality_sorted_indices_original
        self.rank_indices = self.rank_indices_original
        self.frame_ranks_max_index = self.frame_ranks_max_index_original
        self.frame_ranks_max_value = self.frame_ranks_max_value_original

    def set_index_translation(self, index_translation):
        """
        After frames have been marked to be excluded from the further workflow, update the ranking
        tables, based on the index translation list from the frames module.

        :param index_translation: List with indices. For each index in the reduced list of frames
                                  it gives the corresponding index in the original frame list.
        :return: -
        """

        # Set the number of ranks to the number of included frames.
        self.number = len(index_translation)

        self.frame_ranks = [self.frame_ranks_original[index] for index in index_translation]

        # Sort the frame indices in descending order of quality.
        self.quality_sorted_indices = sorted(range(self.number),
                                             key=self.frame_ranks.__getitem__, reverse=True)

        # Compute the inverse index list: For each frame the rank_index is the corresponding index
        # in the sorted frame_ranks list.
        self.rank_indices = [self.quality_sorted_indices.index(index) for index in
                             range(self.number)]

        if self.progress_signal is not None:
            self.progress_signal.emit("Rank all frames", 100)

        # Set the index of the best frame, and normalize all quality values.
        self.frame_ranks_max_index = self.quality_sorted_indices[0]
        self.frame_ranks_max_value = self.frame_ranks[self.frame_ranks_max_index]
        self.frame_ranks /= self.frame_ranks_max_value

    def reset_index_translation(self):
        """
        De-activate index translation and re-establish the original frame ranking data.

        :return: -
        """

        self.number = self.number_original
        self.frame_ranks = self.frame_ranks_original
        self.quality_sorted_indices = self.quality_sorted_indices_original
        self.rank_indices = self.rank_indices_original
        self.frame_ranks_max_index = self.frame_ranks_max_index_original
        self.frame_ranks_max_value = self.frame_ranks_max_value_original

    def find_best_frames(self, number_frames, region_size):
        """
        Find the indices of the best "number_frames" frames under the condition that all indices
        are within an interval of size "region_size".

        :param number_frames: Number of best frames the indices of which are to be found.
        :param region_size: Maximal width of index interval.
        :return: (List of frame indices, quality loss, time line position) with:
                List of frame indices: Indices of frames participating in mean frame computation.
                quality loss: Loss in average frame quality due to range restriction (%).
                time line position: Position of the average frame index relative to the total
                                    duration of the video.
        """

        # Check input arguments for validity.
        if number_frames > region_size:
            raise ArgumentError("Attempt to find " + str(number_frames) + " good frames in "
                                "an index interval of size " + str(region_size))
        elif region_size > self.number:
            raise ArgumentError("Size of best frames region " + str(region_size) + " larger "
                                "than the total number of frames " + str(self.number))

        best_indices = []
        rank_sum_opt = 0.

        # Construct a sliding window on the full index range. For each window position find the
        # best "number_frames" frames. Find the window and the best frame set within with the
        # highest overall score.
        for start_index in range(self.number - region_size + 1):
            end_index = start_index + region_size
            best_indices_in_range = sorted(range(start_index, end_index),
                                           key=self.frame_ranks.__getitem__, reverse=True)[
                                    :number_frames]
            rank_sum = sum([self.frame_ranks[i] for i in best_indices_in_range])
            if rank_sum > rank_sum_opt:
                rank_sum_opt = rank_sum
                best_indices = best_indices_in_range

        # Compare the average frame quality with the optimal choice if no time restrictions were
        # present.
        rank_sum_global = sum(
            [self.frame_ranks[i] for i in self.quality_sorted_indices[:number_frames]])
        quality_loss_percent = round(100. * (rank_sum_global - rank_sum_opt) / rank_sum_global, 1)

        # For the frames included in mean frame computation compute the average position on the
        # video time line.
        cog_mean_frame = round(100 * mean(best_indices) / self.number, 1)

        return best_indices, quality_loss_percent, cog_mean_frame


if __name__ == "__main__":

    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        # names = glob.glob('Images/2012*.tif')
        # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')
        # names = glob.glob('Images/Example-3*.jpg')
        names = glob('Images/Mond_*.jpg')
    else:
        names = 'Videos/another_short_video.avi'
        # names = "E:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2018-03-24\Moon_Tile-024_043939.avi"
        # names = 'Videos/Moon_Tile-024_043939.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()

    # Rank the frames by their overall local contrast.
    start = time()
    rank_frames = RankFrames(frames, configuration)
    rank_frames.frame_score()
    end = time()
    print('Elapsed time in ranking all frames: {}'.format(end - start))

    # for rank, index in enumerate(rank_frames.quality_sorted_indices):
    #     frame_quality = rank_frames.frame_ranks[index]
    #     print("Rank: " + str(rank) + ", Frame no. " + str(index) + ", quality: " + str(frame_quality))
    # for index, frame_quality in enumerate(rank_frames.frame_ranks):
    #     rank = rank_frames.quality_sorted_indices.index(index)
    #     print("Frame no. " + str(index) + ", Rank: " + str(rank) + ", quality: " +
    #           str(frame_quality))

    print("")
    num_frames = len(rank_frames.frame_ranks)
    frame_percent = 10
    num_frames_stacked = max(1, round(num_frames*frame_percent/100.))
    print("Percent of frames to be stacked: ", str(frame_percent), ", numnber: "
           + str(num_frames_stacked))
    quality_cutoff = rank_frames.frame_ranks[rank_frames.quality_sorted_indices[num_frames_stacked]]
    print("Quality cutoff: ", str(quality_cutoff))

    # Plot the frame qualities in chronological order.
    ax1 = plt.subplot(211)

    x = array(rank_frames.frame_ranks)
    plt.ylabel('Frame number')
    plt.gca().invert_yaxis()
    y = array(range(num_frames))
    x_cutoff = full((num_frames,), quality_cutoff)
    plt.xlabel('Quality')
    line1, = plt.plot(x, y, lw=1)
    line2, = plt.plot(x_cutoff, y, lw=1)
    index = 37
    plt.scatter(x[index], y[index], s=20)
    plt.grid(True)

    # Plot the frame qualities ordered by value.
    ax2 = plt.subplot(212)

    x = array([rank_frames.frame_ranks[i] for i in rank_frames.quality_sorted_indices])
    plt.ylabel('Frame rank')
    plt.gca().invert_yaxis()
    y = array(range(num_frames))
    y_cutoff = full((num_frames,), num_frames_stacked)
    plt.xlabel('Quality')
    line3, = plt.plot(x, y, lw=1)
    line4, = plt.plot(x, y_cutoff, lw=1)
    index = 37
    plt.scatter(x[index], y[index], s=20)
    plt.grid(True)

    plt.show()

    number = 3
    window = 5
    start = time()
    best_indices, quality_loss_percent, cog_mean_frame = rank_frames.find_best_frames(number, window)
    end = time()
    print ("\nIndices of best frames in window of size " + str(window) + " found in " +
           str(end - start) + " seconds: " + str(best_indices) +
           "\nQuality loss as compared to unrestricted selection: " +
           str(quality_loss_percent) + "%\nPosition of mean frame in video time line: " +
           str(cog_mean_frame) + "%")



from ctypes import CDLL, byref, c_int
from time import time

from numpy import random, matmul


def mkl_set_num_threads(cores):
    mkl_rt.mkl_set_num_threads(byref(c_int(cores)))


mkl_rt_path = "C:\Python35\Lib\site-packages\\numpy\core\mkl_rt.dll"
# mkl_rt_path = "C:\Python36\Library\\bin\mkl_rt.dll"
mkl_rt = CDLL(mkl_rt_path)

ndim = 20000
iterations = 10

for number_threads in [1, mkl_rt.mkl_get_max_threads()]:
    mkl_set_num_threads(number_threads)
    print("\nNumber of threads used by mkl: " + str(number_threads))

    start = time()
    matrix = random.rand(ndim, ndim)
    vector = random.rand(ndim)
    end = time()
    print("Time for random number generation: " + str(end - start) + " seconds.")

    start = time()
    for iter in range(iterations):
        vector = matmul(matrix, vector)
    end = time()
    print("Time for " + str(iterations) + " matrix multiplies: " + str(end - start) + " seconds.")



from numpy import array, arange

from miscellaneous import Miscellaneous


def evaluate(coefficients, y, x):
    a_f, b_f, c_f, d_f, e_f, g_f = coefficients
    value = a_f * x ** 2 + b_f * y ** 2 + c_f * x * y + d_f * x + e_f * y + g_f
    print ("Function evaluation, y: " + str(y) + ", x: " +str(x) + ", value: " + str(value))
    return value

def create_function_values(coefficients):
    function_values = []
    for y in range(-1,2):
        for x in range(-1,2):
            function_values.append(evaluate(coefficients, y, x))
    return array(function_values)

def evaluate_around_optimum(coefficients, y_opt, x_opt):
    for y in arange(y_opt - 0.1, y_opt + 0.19, 0.1):
        for x in arange(x_opt - 0.1, x_opt + 0.19, 0.1):
            evaluate(coefficients, y, x)

def evaluate_derivatives(coefficients, y, x):
    a_f, b_f, c_f, d_f, e_f, g_f = coefficients
    df_dy = 2. * b_f * y + c_f * x + e_f
    df_dx = 2. * a_f * x + c_f * y + d_f
    print ("derivative_y: " + str(df_dy) + ", derivative_x: " + str(df_dx))


coeff = (1.5, 1.2, 2.5, -0.5, 0.7, 3.)
values = create_function_values(coeff)
print ("\nFunction values: " + str(values))

y, x = Miscellaneous.sub_pixel_solve(values)
evaluate(coeff, y, x)

y_old, x_old = Miscellaneous.sub_pixel_solve_old(values)
print ("Old solution, y: " + str(y_old) + ", x: " + str(x_old))

evaluate_around_optimum(coeff, y_old, x_old)

evaluate_derivatives(coeff, y_old, x_old)



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

import ctypes
import glob
import sys
import os
import platform
import traceback
from ctypes import CDLL

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from skimage import img_as_ubyte
from PyQt5 import QtWidgets

from align_frames import AlignFrames
from alignment_points import AlignmentPoints
from alignment_point_editor import AlignmentPointEditorWidget
from configuration import Configuration
from exceptions import NotSupportedError, InternalError, Error
from frames import Frames
from rank_frames import RankFrames
from stack_frames import StackFrames
from timer import timer

def workflow(input_name, input_type='video', roi=None, automatic_ap_creation=True):
    """
    Execute the whole stacking workflow for a test case. This can either use a video file (.avi .mov .mp4 .ser)
    or still images stored in a single directory.

    :param input_name: Video file (.avi .mov .mp4 .ser) or name of a directory containing still images
    :param input_type: Either "video" or "image" (see "input_name")
    :param roi: If specified, tuple (y_low, y_high, x_low, x_high) with pixel bounds for "region
                of interest"
    :return: average, [average_roi,] color_image_with_aps, stacked_image
             with: - average: global mean frame
                   - average_roi: mean frame restricted to ROI (only if roi is specified)
                   - color_image_with_aps: mean frame overlaid with alignment points and their
                                           boxes (white) and patches (green)
    """

    # Initalize the timer object used to measure execution times of program sections.
    my_timer = timer()
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.

    # For video file input, the Frames constructor expects the video file name for "names".
    if input_type == 'video':
        names = input_name
    # For single image input, the Frames constructor expects a list of image file names for "names".
    else:
        names = [os.path.join(input_name, name) for name in os.listdir(input_name)]
    stacked_image_name = input_name + '.stacked.tiff'

    # The name of the alignment point visualization file is derived from the input video name or
    # the input directory name.
    ap_image_name = input_name + ".aps.tiff"

    print("\n" +
          "*************************************************************************************\n"
          + "Start processing " + str(input_name) +
          "\n*************************************************************************************")
    my_timer.create('Execution over all')

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()

    # Read the frames.
    print("+++ Start reading frames")
    my_timer.create('Read all frames')
    try:
        frames = Frames(configuration, names, type=input_type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + str(e))
        exit()
    my_timer.stop('Read all frames')

    # Rank the frames by their overall local contrast.
    print("+++ Start ranking images")
    my_timer.create('Ranking images')
    rank_frames = RankFrames(frames, configuration)
    rank_frames.frame_score()
    my_timer.stop('Ranking images')
    print("Index of best frame: " + str(rank_frames.frame_ranks_max_index))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        my_timer.create('Select optimal alignment patch')
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)
        my_timer.stop('Select optimal alignment patch')

        print("optimal alignment rectangle, y_low: " + str(y_low_opt) + ", y_high: " +
              str(y_high_opt) + ", x_low: " + str(x_low_opt) + ", x_high: " + str(x_high_opt))

    # Align all frames globally relative to the frame with the highest score.
    print("+++ Start aligning all frames")
    my_timer.create('Global frame alignment')
    try:
        align_frames.align_frames()
    except NotSupportedError as e:
        print("Error: " + e.message)
        exit()
    except InternalError as e:
        print("Warning: " + e.message)
    my_timer.stop('Global frame alignment')

    print("Intersection, y_low: " + str(align_frames.intersection_shape[0][0]) + ", y_high: "
          + str(align_frames.intersection_shape[0][1]) + ", x_low: " \
          + str(align_frames.intersection_shape[1][0]) + ", x_high: " \
          + str(align_frames.intersection_shape[1][1]))

    # Compute the average frame.
    print("+++ Start computing reference frame")
    my_timer.create('Compute reference frame')
    average = align_frames.average_frame()
    my_timer.stop('Compute reference frame')
    print("Reference frame computed from the best " + str(
        align_frames.average_frame_number) + " frames.")

    # If the ROI is to be set to a smaller size than the whole intersection, do so.
    if roi:
        print("+++ Start setting ROI and computing new reference frame")
        my_timer.create('Setting ROI and new reference')
        average_roi = align_frames.set_roi(roi[0], roi[1], roi[2], roi[3])
        my_timer.stop('Setting ROI and new reference')

    # Initialize the AlignmentPoints object.
    my_timer.create('Initialize alignment point object')
    alignment_points = AlignmentPoints(configuration, frames, rank_frames, align_frames)
    my_timer.stop('Initialize alignment point object')

    if automatic_ap_creation:
        # Create alignment points, and create an image with wll alignment point boxes and patches.
        print("+++ Start creating alignment points")
        my_timer.create('Create alignment points')

        # If a ROI is selected, alignment points are created in the ROI window only.
        alignment_points.create_ap_grid()

        my_timer.stop('Create alignment points')
        print("Number of alignment points selected: " + str(len(alignment_points.alignment_points)) +
              ", aps dropped because too dim: " + str(alignment_points.alignment_points_dropped_dim) +
              ", aps dropped because too little structure: " + str(
              alignment_points.alignment_points_dropped_structure))
    else:
        # Open the alignment point editor.
        app = QtWidgets.QApplication(sys.argv)
        alignment_point_editor = AlignmentPointEditorWidget(None, configuration, align_frames,
                                                            alignment_points, None)
        alignment_point_editor.setMinimumSize(800, 600)
        alignment_point_editor.showMaximized()
        app.exec_()

        print("After AP editing, number of APs: " + str(len(alignment_points.alignment_points)))
        count_updates = 0
        for ap in alignment_points.alignment_points:
            if ap['reference_box'] is not None:
                continue
            count_updates += 1
            AlignmentPoints.set_reference_box(ap, alignment_points.mean_frame)
        print("Buffers allocated for " + str(count_updates) + " alignment points.")

    # Produce an overview image showing all alignment points.
    if roi:
        color_image_with_aps = alignment_points.show_alignment_points(average_roi)
    else:
        color_image_with_aps = alignment_points.show_alignment_points(average)

    # For each alignment point rank frames by their quality.
    my_timer.create('Rank frames at alignment points')
    print("+++ Start ranking frames at alignment points")
    alignment_points.compute_frame_qualities()
    my_timer.stop('Rank frames at alignment points')

    # Allocate StackFrames object.
    stack_frames = StackFrames(configuration, frames, rank_frames, align_frames, alignment_points, my_timer)

    # Stack all frames.
    print("+++ Start stacking frames")
    stack_frames.stack_frames()

    # Merge the stacked alignment point buffers into a single image.
    print("+++ Start merging alignment patches")
    stacked_image = stack_frames.merge_alignment_point_buffers()

    # If the drizzle factor is 1.5, reduce the pixel resolution of the stacked image buffer
    # to half the size used in stacking.
    if configuration.drizzle_factor_is_1_5:
        print("+++ Start reducing image buffer size")
        stack_frames.half_stacked_image_buffer_resolution()

    # Save the stacked image as 16bit int (color or mono).
    my_timer.create('Saving the final image')
    Frames.save_image(stacked_image_name, stacked_image, color=frames.color,
                      header=configuration.global_parameters_version)
    my_timer.stop('Saving the final image')

    # Print out timer results.
    my_timer.stop('Execution over all')
    my_timer.print()

    # Write the image with alignment points.
    Frames.save_image(ap_image_name, color_image_with_aps, color=True,
                      header=configuration.global_parameters_version)

    # If a ROI is selected, return both the original and the reduced-size average frame.
    if roi:
        return average, average_roi, color_image_with_aps, stacked_image
    else:
        return average, color_image_with_aps, stacked_image


if __name__ == "__main__":
    """
    This File contains a test main program. It goes through the whole process without using a
    graphical unser interface. It is not intended to be used in production runs.
    """

    ####################################### Specify test case ######################################
    redirect_stdout = False
    show_results = True
    input_type = 'video'
    # input_directory = 'D:/SW-Development/Python/PlanetarySystemStacker/Examples/Moon_2018-03-24'
    input_directory = 'D:/SW-Development/Python/PlanetarySystemStacker/Examples/Jupiter_short'
    # input_type = 'image'
    # input_directory = 'D:/SW-Development/Python/PlanetarySystemStacker/Examples/Moon_2011-04-10'
    automatic_ap_creation = False
    roi = None
    # roi = (400, 700, 300, 800)
    ####################################### Specify test case end ##################################

    # Redirect standard output to a file if requested.
    protocol_file = None
    if redirect_stdout:
        print("Redirecting stdout, please check Protocol.txt file in input directory.",
              file=sys.stderr)
        stdout_saved = sys.stdout
        protocol_file = open(os.path.join(input_directory, 'Protocol.txt'), 'a')
        sys.stdout = protocol_file

    # For videos, batch processing is done for all videos in the input directory.
    if input_type == 'video':
        input_names = glob.glob(os.path.join(input_directory, '*.avi'))
    # For images, it is assumed that the input directory contains one or several directories with
    # image files.
    else:
        if input_type != 'image':
            print("WARNING: Wrong input spec, assuming 'image'")
        input_directory_content = os.listdir(input_directory)
        # input_names = [dir for dir in input_directory_content if os.path.isdir(dir)]
        input_names = []
        for dir in input_directory_content:
            dir_abs = os.path.join(input_directory, dir)
            if os.path.isdir(dir_abs):
                input_names.append(dir_abs)

    # Feedback on number of files found to process
    if len(input_names) > 0:
        print("+++ Start processing", len(input_names), input_type)
    else:
        print("+++ No files found to process")

    # Start the processing workflow in batch mode for all AVIs / file directories.
    try:
        for input_name in input_names:
            if roi:
                average, average_roi, color_image_with_aps, stacked_image = workflow(input_name,
                    input_type=input_type, roi=roi,
                    automatic_ap_creation=automatic_ap_creation)
            else:
                average, color_image_with_aps, stacked_image = workflow(input_name,
                    input_type=input_type,
                    automatic_ap_creation=automatic_ap_creation)

            # Interrupt the workflow to display resulting images only if requested.
            if show_results:
                # Show the full average frame.
                plt.imshow(average, cmap='Greys_r')
                plt.show()

                if roi:
                    # Show the ROI average frame.
                    plt.imshow(average_roi, cmap='Greys_r')
                    plt.show()

                # Show alignment points and patches
                plt.imshow(color_image_with_aps)
                plt.show()

                # Convert the stacked image to 8bit and show in Window.
                plt.imshow(img_as_ubyte(stacked_image))
                plt.show()
    except:
        exec_info = sys.exc_info()
        if protocol_file is None:
            traceback.print_tb(exec_info[2])
        else:
            traceback.print_tb(exec_info[2], file=protocol_file)
        print(exec_info[1])

    # Redirect stdout back to normal.
    if redirect_stdout:
        sys.stdout = stdout_saved
        protocol_file.close()



from math import ceil

from numpy import ndarray, zeros

from miscellaneous import Miscellaneous


def compute_bounds(x_low, x_high, drizzle_factor, shift):
    """
    Compute the index bounds in the original (non-drizzled) frame from where the AP patch is to be
    copied, and the (drizzled) index bounds in the summation buffer. The data are copied with stride
     "drizzle_factor".

    :param x_low: Lower index bound of AP patch in original frame.
    :param x_high: Upper index bound of AP patch in original frame.
    :param drizzle_factor: (integer) drizzle factor, typically 2 or 3.
    :param shift: Shift (float) between AP patches in summation buffer and original frame,
                  expressed in non-drizzled units.
    :return: Tuple (x_low_from, x_high_from, xd_low_to, xd_high_to, x_offset) with:
             x_low_from: Lower source index bound for copy operation in original frame.
             x_high_from: Upper source index bound for copy operation in original frame (stride 1).
             xd_low_to: Lower target index bound for copy operation in summation buffer.
             xd_high_to: Upper target index bound for copy operation in summation buffer
                         (stride "drizzle_factor").
             x_offset: Index offset in target patch for begin of copy operation
                       (0, 1, ..., drizzle_factor-1).
    """

    # Compute integer shift in drizzled target grid closest to given shift (float).
    shift_d = int(round(drizzle_factor * shift))

    # Translate into original index coordinates (not integer any more).
    shift_rounded = shift_d / drizzle_factor

    # If the shift stays in the original grid, the offset in the drizzled grid patch is zero.
    if shift_rounded.is_integer():
        shift_rounded = int(shift_rounded)
        x_low_from = x_low + shift_rounded
        x_high_from = x_high + shift_rounded
        x_offset = 0

    # Otherwise the target indices in the drizzled grid patch start at a non-zero offset.
    else:
        shift_ceil = ceil(shift)
        x_low_from = x_low + shift_ceil
        x_high_from = x_high + shift_ceil
        x_offset = int(drizzle_factor * shift_ceil - shift_d)

    xd_low_to = drizzle_factor * x_low_from - shift_d
    xd_high_to = drizzle_factor * x_high_from - shift_d

    return x_low_from, x_high_from, xd_low_to, xd_high_to, x_offset


def compute_bounds_2d(y_low, y_high, x_low, x_high, shift_y, shift_x, drizzle_factor):
    """
    Compute the index bounds in the original (non-drizzled) frame from where the AP patch is
    to be copied, and the (drizzled) index bounds in the summation buffer. The data are copied with
    stride "drizzle_factor".

    :param y_low: Lower y index bound of AP patch in original frame.
    :param y_high: Upper y index bound of AP patch in original frame.
    :param x_low: Lower x index bound of AP patch in original frame.
    :param x_high: Upper x index bound of AP patch in original frame.
    :param shift_y: Shift (float) in y between AP patches in summation buffer and original frame,
                    expressed in non-drizzled units.
    :param shift_x: Shift (float) in x between AP patches in summation buffer and original frame,
                    expressed in non-drizzled units.
    :param drizzle_factor: (integer) drizzle factor, typically 2 or 3.
    :return: Tuple (y_low_from, y_high_from, x_low_from, x_high_from, y_offset, x_offset) with:
             y_low_from: Lower source y index bound for copy operation in original frame.
             y_high_from: Upper source y index bound for copy operation in original frame (
             stride 1).
             x_low_from: Lower source x index bound for copy operation in original frame.
             x_high_from: Upper source x index bound for copy operation in original frame (
             stride 1).
             y_offset: Index offset in y direction in target patch for begin of copy operation
                       (0, 1, ..., drizzle_factor-1).
             x_offset: Index offset in x direction in target patch for begin of copy operation
                       (0, 1, ..., drizzle_factor-1).
    """

    # Compute integer shift in both directions in drizzled target grid closest to given shift
    # (float).
    shift_d_y = int(round(drizzle_factor * shift_y))
    shift_d_x = int(round(drizzle_factor * shift_x))

    # Translate into original index coordinates (not integer any more).
    shift_rounded_y = shift_d_y / drizzle_factor
    shift_rounded_x = shift_d_x / drizzle_factor

    # If the shift stays in the original grid, the offset in the drizzled grid patch is zero.
    if shift_rounded_y.is_integer():
        shift_rounded_y = int(shift_rounded_y)
        y_low_from = y_low + shift_rounded_y
        y_high_from = y_high + shift_rounded_y
        y_offset = 0

    # Otherwise the target indices in the drizzled grid patch start at a non-zero offset.
    else:
        shift_ceil = ceil(shift_y)
        y_low_from = y_low + shift_ceil
        y_high_from = y_high + shift_ceil
        y_offset = int(drizzle_factor * shift_ceil - shift_d_y)

    # Do the same for the x coordinate direction.
    if shift_rounded_x.is_integer():
        shift_rounded_x = int(shift_rounded_x)
        x_low_from = x_low + shift_rounded_x
        x_high_from = x_high + shift_rounded_x
        x_offset = 0

    # Otherwise the target indices in the drizzled grid patch start at a non-zero offset.
    else:
        shift_ceil = ceil(shift_x)
        x_low_from = x_low + shift_ceil
        x_high_from = x_high + shift_ceil
        x_offset = int(drizzle_factor * shift_ceil - shift_d_x)

    return y_low_from, y_high_from, x_low_from, x_high_from, y_offset, x_offset

def remap_rigid_drizzled(frame, buffer, offset_counters, shift_y, shift_x, y_low, y_high, x_low,
                         x_high, drizzle_factor):
    """
    The alignment point patch is taken from the given frame with a constant shift in x and y
    directions. The shifted patch is then added to the given alignment point buffer.

    Please note that all indices related to the frame, and all input arguments in particular, are
    in non-drizzled coordinates. The size of the buffer, on the other side, is extended by the
    drizzling factor "drizzle_factor". To enable sub-pixel accuracy, shift values are input as
    non-integer floats.

    :param frame: frame to be stacked
    :param buffer: Stacking buffer of the corresponding alignment point (extended for drizzling)
    :param offset_counters: Integer array, size drizzle_factor x drizzle_factor, with contribution
                            counts at all drizzle locations.
    :param shift_y: Constant shift in y direction between frame stack and current frame (float)
    :param shift_x: Constant shift in x direction between frame stack and current frame (float)
    :param y_low: Lower y index of the quality window on which this method operates
    :param y_high: Upper y index of the quality window on which this method operates
    :param x_low: Lower x index of the quality window on which this method operates
    :param x_high: Upper x index of the quality window on which this method operates
    :param drizzle_factor: (integer) drizzle factor, typically 2 or 3.
    :return: -
    """

    # Compute integer shift in both directions in drizzled target grid closest to given shift
    # (float).
    shift_d_y = int(round(drizzle_factor * shift_y))
    shift_d_x = int(round(drizzle_factor * shift_x))

    # Translate into original index coordinates (not integer any more).
    shift_rounded_y = shift_d_y / drizzle_factor
    shift_rounded_x = shift_d_x / drizzle_factor

    # If the shift stays in the original grid, the offset in the drizzled grid patch is zero.
    if shift_rounded_y.is_integer():
        shift_rounded_y = int(shift_rounded_y)
        y_low_from = y_low + shift_rounded_y
        y_high_from = y_high + shift_rounded_y
        y_offset = 0

    # Otherwise the target indices in the drizzled grid patch start at a non-zero offset.
    else:
        shift_ceil = ceil(shift_y)
        y_low_from = y_low + shift_ceil
        y_high_from = y_high + shift_ceil
        y_offset = int(drizzle_factor * shift_ceil - shift_d_y)

    # Do the same for the x coordinate direction.
    if shift_rounded_x.is_integer():
        shift_rounded_x = int(shift_rounded_x)
        x_low_from = x_low + shift_rounded_x
        x_high_from = x_high + shift_rounded_x
        x_offset = 0

    # Otherwise the target indices in the drizzled grid patch start at a non-zero offset.
    else:
        shift_ceil = ceil(shift_x)
        x_low_from = x_low + shift_ceil
        x_high_from = x_high + shift_ceil
        x_offset = int(drizzle_factor * shift_ceil - shift_d_x)

    # Compute index bounds for "source" patch in current frame, and for summation buffer
    # ("target"). Because of local warp effects, the indexing may reach beyond frame borders.
    # In this case reduce the copy area.
    frame_size_y = frame.shape[0]
    y_low_target = y_offset
    # If the shift reaches beyond the frame, reduce the copy area.
    if y_low_from < 0:
        y_low_target -= y_low_from*drizzle_factor
        y_low_from = 0
    if y_high_from > frame_size_y:
        y_high_from = frame_size_y
    y_high_target = y_low_target + (y_high_from - y_low_from)*drizzle_factor

    # The same in x direction.
    frame_size_x = frame.shape[1]
    x_low_target = x_offset
    # If the shift reaches beyond the frame, reduce the copy area.
    if x_low_from < 0:
        x_low_target -= x_low_from * drizzle_factor
        x_low_from = 0
    if x_high_from > frame_size_x:
        x_high_from = frame_size_x
    x_high_target = x_low_target + (x_high_from - x_low_from) * drizzle_factor

    # Add the shifted frame patch to the AP buffer.
    buffer[y_low_target:y_high_target:drizzle_factor, x_low_target:x_high_target:drizzle_factor] += \
        frame[y_low_from:y_high_from, x_low_from:x_high_from]

    # Increment the offset counter for the current drizzle position.
    offset_counters[y_offset, x_offset] += 1

def equalize_ap_patch(patch, offset_counters, stack_size, drizzle_factor):
    """
    During drizzling the AP patch gets different numbers of frame contributions at different
    positions in the drizzling pattern, depending on the distribution of sup-pixel shift values.
    The array "offset_counters" contains these numbers, with the sum of all those numbers being the
    total stack size.

    Scale all patch entries with "total stack size" / "offset counter". The result are uniform
    patch entries which look as if there had been "total stack size" contributions in all drizzle
    pattern locations.

    Special care has to be taken at locations where not a single frame contributed (called "holes").
    Here insert the average of patch values at all drizzle positions over a circle with minimum
    radius containing at least one non-zero contribution.

    :param patch: AP buffer after stacking with drizzling.
    :param offset_counters: Integer array, size drizzle_factor x drizzle_factor, with contribution
                            counts at all drizzle locations.
    :param stack_size: Overall number of frames stacked
    :param drizzle_factor: Factor by which the number of pixels is multiplied in each direction.
    :return: Number of drizzle locations without frame contributions.
    """

    dim_y, dim_x = patch.shape[:2]
    holes = []

    # First normalize the patch locations with non-zero contributions.
    for y_offset in range(drizzle_factor):
        for x_offset in range(drizzle_factor):
            if offset_counters[y_offset, x_offset]:
                normalization_factor = stack_size / offset_counters[y_offset, x_offset]
                patch[y_offset:dim_y:drizzle_factor,
                x_offset:dim_x:drizzle_factor] *= normalization_factor
            # For locations with zero contributions (holes) remember the location.
            else:
                holes.append((y_offset, x_offset))

    # Now fill the holes with interpolated values from locations close by.
    for (y_offset, x_offset) in holes:

        # Initialize the buffer locations witz zeros.
        patch[y_offset:dim_y:drizzle_factor, x_offset:dim_x:drizzle_factor] = 0.

        # Look for the circle with smallest radius around the hole with at least one non-zero
        # contribution.
        for radius in range(1, drizzle_factor):
            n_success = 0
            for (y, x) in Miscellaneous.circle_around(y_offset, x_offset, radius):
                if 0 <= y < drizzle_factor and 0 <= x < drizzle_factor and (y, x) not in holes:
                    # A non-zero entry is found, add its contribution to the buffer.
                    patch[y_offset:dim_y:drizzle_factor, x_offset:dim_x:drizzle_factor] += \
                        patch[y:dim_y:drizzle_factor, x:dim_x:drizzle_factor]
                    n_success += 1

            # There was at least one non-zero contribution on the circle with this radius. Normalize
            # the buffer with the number of contributions and continue with the next hole.
            if n_success:
                patch[y_offset:dim_y:drizzle_factor,
                x_offset:dim_x:drizzle_factor] *= 1. / n_success
                break

    # Return the number of holes in the drizzle pattern.
    return len(holes)


def test_index_computations():
    # Set parameters and example shift value.
    drizzle_factor = 3
    x_low = 4
    x_high = 6
    shift = -2.52

    # Compute the index bounds and offset.
    x_low_from, x_high_from, xd_low_to, xd_high_to, x_offset = compute_bounds(x_low, x_high,
                                                                              drizzle_factor,
                                                                              shift)

    # Print the results.
    print("x_low_from: " + str(x_low_from) + ", x_high_from: " + str(
        x_high_from) + "\nxd_low_to: " + str(xd_low_to) + ", xd_high_to: " + str(
        xd_high_to) + "\nx_offset: " + str(x_offset))


def test_index_computations_2d():
    # Set parameters and example shift value.
    drizzle_factor = 3
    y_low = 4
    y_high = 6
    shift_y = 3.14
    x_low = 4
    x_high = 6
    shift_x = -2.52

    # Compute the index bounds and offset.
    y_low_from, y_high_from, x_low_from, x_high_from, y_offset, x_offset = \
        compute_bounds_2d(y_low, y_high, x_low, x_high, shift_y, shift_x, drizzle_factor)

    # Print the results.
    print("y_low_from: " + str(y_low_from) + ", y_high_from: " + str(
        y_high_from) + "\ny_offset: " + str(y_offset) + "\nx_low_from: " + str(
        x_low_from) + ", x_high_from: " + str(x_high_from) + "\nx_offset: " + str(x_offset))


def test_remap_rigid_drizzled():
    drizzle_factor = 3
    y_dim = 20
    x_dim = 16
    y_dim_patch = 3
    x_dim_patch = 4
    frame = ndarray(shape=(y_dim, x_dim), dtype=int)
    for y in range(y_dim):
        for x in range(x_dim):
            frame[y, x] = 1000 * y + x
    patch = zeros((y_dim_patch * drizzle_factor, x_dim_patch * drizzle_factor), dtype=int)
    offset_counters = zeros(shape=(drizzle_factor, drizzle_factor), dtype=int)

    y_low = 10
    y_high = y_low + y_dim_patch
    shift_y = 1.26
    x_low = 6
    x_high = x_low + x_dim_patch
    shift_x = -1.6

    y_low_from, y_high_from, x_low_from, x_high_from, y_offset, x_offset = \
        compute_bounds_2d(y_low, y_high, x_low, x_high, shift_y, shift_x, drizzle_factor)

    print("y_low_from: " + str(y_low_from) + ", y_high_from: " + str(
        y_high_from) + "\ny_offset: " + str(y_offset) + "\nx_low_from: " + str(
        x_low_from) + ", x_high_from: " + str(x_high_from) + "\nx_offset: " + str(x_offset))

    remap_rigid_drizzled(frame, patch, offset_counters, shift_y, shift_x, y_low, y_high, x_low,
                         x_high, drizzle_factor)

    print("patch: " + str(patch))


def test_equalize_ap_patch():
    drizzle_factor = 3
    y_dim = 3
    x_dim = 4
    y_dim_patch = y_dim * drizzle_factor
    x_dim_patch = x_dim * drizzle_factor
    patch = zeros(shape=(y_dim_patch, x_dim_patch), dtype=float)
    offset_counters = zeros(shape=(drizzle_factor, drizzle_factor), dtype=int)
    offset_counters[0, 0] = 1
    offset_counters[0, 1] = 2
    offset_counters[1, 0] = 1
    offset_counters[2, 2] = 3
    stack_size = offset_counters.sum()
    for y_offset in range(drizzle_factor):
        for x_offset in range(drizzle_factor):
            counter = offset_counters[y_offset, x_offset]
            if counter:
                for y in range(y_offset, y_dim_patch, drizzle_factor):
                    for x in range(x_offset, x_dim_patch, drizzle_factor):
                        patch[y, x] = counter * (1000 * y + x)
                        # patch[y, x] = counter

    print("patch before equalization: \n" + str(patch))
    len_holes = equalize_ap_patch(patch, offset_counters, stack_size, drizzle_factor)
    print("patch after equalization: \n" + str(patch))
    print("Number of holes in drizzle pattern: " + str(len_holes))

def test_numpy():
    a = ndarray((4,5,3), dtype=int)
    for y in range(4):
        for x in range(5):
            for c in range(3):
                a[y, x, c] = y*100 + x*10 + c
    print("a: " + str(a))

    b = a[2:3, 2:5]
    print ("b: " + str(b))

    d = a[2:3, 2:5, :]
    print("d: " + str(d))


# Main program: Control the test to be performed.
# test_index_computations()
# test_index_computations_2d()
# test_remap_rigid_drizzled()
test_equalize_ap_patch()
# test_numpy()



from os import listdir, rename, remove
from os.path import splitext, join, dirname

# The following lists define the allowed file extensions for still images and videos.
image_extensions = ['.tif', '.tiff', '.fit', '.fits', '.jpg', '.png']

dir_name = "D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2011-04-10\South"
names = [join(dir_name, name) for name in listdir(dir_name) if splitext(name)[-1].lower() in image_extensions]

print ("names: " + str(names))

